#!/usr/bin/python

'''
This script contains helper functions to parse percolator in and output files (tab delimited formats)
'''

from __future__ import print_function

import sys
import csv
import collections

from .. import parsers

PercolatorPoutPsmsBase = collections.namedtuple("PercolatorPoutPsms", "id filename scannr charge svm_score qvalue PEP peptide proteins")

class PercolatorPoutPsms(PercolatorPoutPsmsBase):
  def toList(self):
    return [self.id, self.svm_score, self.qvalue, self.PEP, self.peptide] + self.proteins
  
  def toString(self):
    return "\t".join(map(str, self.toList()))
  
# works for peptide and psms pouts
def parsePsmsPout(poutFile, qThresh = 1.0, proteinMap = None, parseId = True, fixScannr = False):
  reader = parsers.getTsvReader(poutFile)
  headers = next(reader) # save the header
  
  cruxOutput = True if "percolator score" in headers else False
  if cruxOutput:
    proteinCol = headers.index('protein id')
    fileIdxCol = headers.index('file_idx')
    scanCol = headers.index('scan')
    chargeCol = headers.index('charge')
    scoreCol = headers.index('percolator score')
    qvalCol = headers.index('percolator q-value')
    peptideCol = headers.index('sequence')
    terminalsCol = headers.index('flanking aa')
  else:
    proteinCol = headers.index('proteinIds')
    qvalCol = headers.index('q-value')
  
  fixScannr = "_msfragger" in poutFile or "_moda" in poutFile or "_msgf" in poutFile or fixScannr
  for row in reader:
    if float(row[qvalCol]) <= qThresh:
      if cruxOutput:
        proteins = list(set(row[proteinCol].split(',')))
      else:
        proteins = row[5:]
      
      if proteinMap:
        proteins = list(map(proteinMap, proteins))
      
      if cruxOutput:
        yield PercolatorPoutPsms(row[fileIdxCol] + "_" + row[scanCol] + "_" + row[chargeCol], int(row[fileIdxCol]), int(row[scanCol]), int(row[chargeCol]), float(row[scoreCol]), float(row[qvalCol]), 1.0, row[terminalsCol][0] + "." + row[peptideCol] + "." + row[terminalsCol][1], proteins)
      elif parseId:
        yield PercolatorPoutPsms(row[0], getFileName(row[0], fixScannr), getId(row[0], fixScannr), getCharge(row[0]), float(row[1]), float(row[qvalCol]), float(row[3]), row[4], proteins)
      else:
        yield PercolatorPoutPsms(row[0], "", 0, 0, float(row[1]), float(row[2]), float(row[3]), row[4], proteins)
    else:
      break

def toList(psm):
  l = list(psm)
  return l[:-1] + l[-1]
  
def getId(PSMid, msgf = False):
  if msgf:
    return int((PSMid.split('_'))[-3]) / 100
  else:
    return int((PSMid.split('_'))[-3])

def getCharge(PSMid):
  return int((PSMid.split('_'))[-2])

def getFileName(PSMid, msgf = False):
  if msgf:
    return '_'.join(PSMid.split('_')[:-6])
  else:
    return '_'.join(PSMid.split('_')[:-3])

if __name__ == "__main__":
   main(sys.argv[1:])



#!/usr/bin/python

from __future__ import print_function

import os
import sys
import warnings

from ..triqler import __version__, __copyright__
from .. import hyperparameters
from .. import parsers

def main():
  print('Triqler.distribution.plot_hyperparameter_fits version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  params['warningFilter'] = "ignore"
  with warnings.catch_warnings():
    warnings.simplefilter(params['warningFilter'])
    plotHyperparameterFits(args.in_file, params)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  apars.add_argument('in_file', default=None, metavar = "IN_FILE",
                     help='''Triqler input file.
                          ''')
  
  apars.add_argument('--decoy_pattern', default = "decoy_", metavar='P', 
                     help='Prefix for decoy proteins.')
  
  apars.add_argument('--no_plots',
                     help='Only print out hyperparameter estimates, without plotting the empirical and fitted distributions.',
                     action='store_true')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['decoyPattern'] = args.decoy_pattern
  params['skipPlots'] = args.no_plots
    
  return args, params
  
def plotHyperparameterFits(triqlerInputFile, params):  
  if not os.path.isfile(triqlerInputFile):
    sys.exit("Could not locate input file %s. Check if the path to the input file is correct." % triqlerInputFile)
  
  peptQuantRowFile = triqlerInputFile + ".pqr.tsv"
  if not os.path.isfile(peptQuantRowFile):
    sys.exit("Could not locate peptide quantification file %s. Run triqler to generate this file." % peptQuantRowFile)
  
  _, params['groups'], _, peptQuantRows = parsers.parsePeptideQuantFile(peptQuantRowFile)
  
  if not params['skipPlots']:
    import matplotlib
    matplotlib.rcParams['axes.unicode_minus'] = False
  
  print("")
  print("N.B. naming of hyperparameters differs from notation in the paper:")
  print("  muDetect = mu_m; sigmaDetect = sigma_m;")
  print("  muXIC = mu_f; sigmaXIC = sigma_f;")
  print("  muProtein = mu_y; sigmaProtein = sigma_y;")
  print("  muFeatureDiff = mu_d; sigmaFeatureDiff = sigma_d;")
  print("  shapeInGroupStdevs = alpha_s; scaleInGroupStdevs = beta_s")
  print("")
  
  hyperparameters.fitPriors(peptQuantRows, params, plot = not params['skipPlots'])
  if not params['skipPlots']:
    import matplotlib.pyplot as plt
    plt.show()
  
if __name__ == "__main__":
  main()



'''
Create Triqler input files by combining Percolator output with the feature 
groups from Dinosaur. Requires Triqler and NumPy to be installed.

If Percolator output is unavailable, one can mimick this format by providing
a tab-delimited file with the following columns (including a header row):

PSMId <tab> score <tab> q-value <tab> posterior_error_prob <tab> peptide <tab> proteinIds

The q-value and posterior_error_prob can be set to 0, as they are not used 
here. Furthermore, the file should be sorted by the score column, where higher 
scores indicate more confident hits and the highest score is on top of the list.
'''

from __future__ import print_function

import os
import sys
import glob
import numpy as np
import collections

from .. import parsers
from ..triqler import __version__, __copyright__

from . import helpers

def main():
  print('Triqler.convert.dinosaur version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  # hack for windows
  if len(args.in_file) == 1 and '*' in args.in_file[0]:
    args.in_file = glob.glob(args.in_file[0])
  
  convertDinosaurToTriqler(args.file_list_file, args.in_file, args.psm_files.split(","), args.out_file, params)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  
  requiredNamed = apars.add_argument_group('required arguments')
  
  apars.add_argument('in_file', default=None, metavar = "IN_FILE", nargs='*',
                     help='''Files containing the mapping of scan numbers to precursor information from the add-dinosaur-precursors script. To easily specify multiple files one can use wildcards, e.g. dinosaur_mapping_files/*.txt
                          ''')
  
  requiredNamed.add_argument('--file_list_file', metavar='L', 
                     help='Simple text file with spectrum file names in first column and condition in second column.',
                     required = True)
  
  requiredNamed.add_argument('--psm_files', metavar='TARGET,DECOY', 
                     help='Percolator PSM output files, separated commas. Both target and decoy output files are needed, with the target file(s) specified first.',
                     required = True)
  
  apars.add_argument('--out_file', default = "triqler_input.tsv", metavar='OUT', 
                     help='''Path to triqler input file (writing in TSV format).
                          ''')
  
  apars.add_argument('--skip_normalization',
                     help='Skip retention-time based intensity normalization.',
                     action='store_true')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['simpleOutputFormat'] = True
  params['skipNormalization'] = args.skip_normalization
  params['plotScatter'] = False
  params['skipMBR'] = True
  
  return args, params
  
def convertDinosaurToTriqler(fileListFile, mappedPrecursorFiles, psmsOutputFiles, triqlerInputFile, params):
  fileInfoList = parsers.parseFileList(fileListFile)
  
  peptideToFeatureMap = parseDinosaurMapFiles(mappedPrecursorFiles, fileInfoList, psmsOutputFiles)
  
  rTimeArrays, factorArrays = helpers.getNormalizationFactorArrays(peptideToFeatureMap, fileInfoList, params)
  
  helpers.writeTriqlerInputFile(triqlerInputFile, peptideToFeatureMap, rTimeArrays, factorArrays, params)

def parseDinosaurMapFiles(mappedPrecursorFiles, fileInfoList, psmsOutputFiles):
  fileList, _, _, _ = zip(*fileInfoList)
  specToPeptideMap = helpers.parsePsmsPoutFiles(psmsOutputFiles, key = lambda psm : (psm.filename, psm.scannr))
  
  fileKey = -1
  linkPEP = 0.0
  peptideToFeatureMap = collections.defaultdict(list)
  for mappedPrecursorFile in mappedPrecursorFiles:
    print("Processing", mappedPrecursorFile)
    fileName = parsers.getMappedPrecursorOriginalFile(mappedPrecursorFile)
    if not fileName in fileList:
      print("Warning: Could not find %s in the specified file list, skipping row" % row[fileCol])
      continue
    
    fileIdx = fileList.index(fileName)
    run, condition, sample, fraction = fileInfoList[fileIdx]
    for spMap in parsers.parseMappedPrecursorFile(mappedPrecursorFile):
      if fileKey == -1:
        if specToPeptideMap((fileName, spMap.scanNr))[0] != helpers.getDefaultPeptideHit()[0]:
          fileKey = 1
        elif specToPeptideMap((fileIdx, spMap.scanNr))[0] != helpers.getDefaultPeptideHit()[0]:
          fileKey = 2
      
      if fileKey == 2:
        fileId = fileIdx # crux percolator
      else:
        fileId = fileName # standalone percolator
      
      peptide, proteins, svmScore, charge = specToPeptideMap((fileId, spMap.scanNr))
      if peptide != helpers.getDefaultPeptideHit()[0]:
        key = (peptide, charge)
        if key in peptideToFeatureMap:
          featureClusterIdx = peptideToFeatureMap[key][0][0].featureClusterId
        else:
          featureClusterIdx = len(peptideToFeatureMap)
        triqlerRow = parsers.TriqlerInputRow(sample, condition, spMap.charge, spMap.scanNr, linkPEP, featureClusterIdx, svmScore, spMap.intensity, peptide, proteins)
        peptideToFeatureMap[key].append((triqlerRow, spMap.rTime, fraction))
      
  if len(peptideToFeatureMap) == 0:
    sys.exit("Error: Could not map any PSM to the appropriate dinosaur feature.")
  return peptideToFeatureMap

if __name__ == "__main__":
   main()



#!/usr/bin/python

from __future__ import print_function

import csv
import sys
import os
import itertools

import numpy as np
from scipy.stats import hypsecant, gamma, norm, binom, t, cauchy
from scipy.optimize import curve_fit

from . import parsers

def fitPriors(peptQuantRows, params, printImputedVals = False, plot = False):
  params['proteinQuantCandidates'] = np.arange(-5.0, 5.0 + 1e-10, 0.01) # log10 of protein ratio  
  qc = params['proteinQuantCandidates']
  params['proteinDiffCandidates'] = np.linspace(2*qc[0], 2*qc[-1], len(qc)*2-1)
  
  protQuantRows = parsers.filterAndGroupPeptides(peptQuantRows, lambda x : not x.protein[0].startswith(params['decoyPattern']))
  
  imputedVals, imputedDiffs, observedXICValues, protQuants, protDiffs, protStdevsInGroup, protGroupDiffs = list(), list(), list(), list(), list(), list(), list()
  quantRowsCollection = list()
  for prot, quantRows in protQuantRows:
    quantRows, quantMatrix = parsers.getQuantMatrix(quantRows)
    
    quantMatrixNormalized = [parsers.geoNormalize(row) for row in quantMatrix]
    quantRowsCollection.append((quantRows, quantMatrix))
    geoAvgQuantRow = getProteinQuant(quantMatrixNormalized, quantRows)
    
    protQuants.extend([np.log10(x) for x in geoAvgQuantRow if not np.isnan(x)])

    args = parsers.getQuantGroups(geoAvgQuantRow, params["groups"], np.log10)
    #means = list()
    for group in args:
      if np.count_nonzero(~np.isnan(group)) > 1:
        protDiffs.extend(group - np.mean(group))
        protStdevsInGroup.append(np.std(group))
      #if np.count_nonzero(~np.isnan(group)) > 0:
      #  means.append(np.mean(group))
    
    #if np.count_nonzero(~np.isnan(means)) > 1:
    #  for mean in means:  
    #    #protGroupDiffs.append(mean - np.mean(means))
    #    protGroupDiffs.append(mean)
    
    quantMatrixFiltered = np.log10(np.array([x for x, y in zip(quantMatrix, quantRows) if y.combinedPEP < 1.0]))  
    observedXICValues.extend(quantMatrixFiltered[~np.isnan(quantMatrixFiltered)])
    
    # counts number of NaNs per run, if there is only 1 non NaN in the column, we cannot use it for estimating the imputedDiffs distribution
    numNonNaNs = np.count_nonzero(~np.isnan(quantMatrixFiltered), axis = 0)[np.newaxis,:]
    xImps = imputeValues(quantMatrixFiltered, geoAvgQuantRow, np.log10(geoAvgQuantRow))
    imputedDiffs.extend((xImps - quantMatrixFiltered)[(~np.isnan(quantMatrixFiltered)) & (np.array(numNonNaNs) > 1)])
    #imputedVals.extend(xImps[(np.isnan(quantMatrixFiltered)) & (np.array(numNonNaNs) > 1)])
  
  fitLogitNormal(observedXICValues, params, plot)
  
  fitDist(protQuants, funcHypsec, "log10(protein ratio)", ["muProtein", "sigmaProtein"], params, plot)
    
  fitDist(imputedDiffs, funcHypsec, "log10(imputed xic / observed xic)", ["muFeatureDiff", "sigmaFeatureDiff"], params, plot)
  
  fitDist(protStdevsInGroup, funcGamma, "stdev log10(protein diff in group)", ["shapeInGroupStdevs", "scaleInGroupStdevs"], params, plot, x = np.arange(-0.1, 1.0, 0.005))
  
  sigmaCandidates = np.arange(0.001, 3.0, 0.001)
  gammaCandidates = funcGamma(sigmaCandidates, params["shapeInGroupStdevs"], params["scaleInGroupStdevs"])
  support = np.where(gammaCandidates > max(gammaCandidates) * 0.01)
  params['sigmaCandidates'] = np.linspace(sigmaCandidates[support[0][0]], sigmaCandidates[support[0][-1]], 20)
  
  params['proteinPrior'] = funcLogHypsec(params['proteinQuantCandidates'], params["muProtein"], params["sigmaProtein"])
  if "shapeInGroupStdevs" in params:
    params['inGroupDiffPrior'] = funcHypsec(params['proteinDiffCandidates'], 0, params['sigmaCandidates'][:, np.newaxis])
  else: # if we have technical replicates, we could use a delta function for the group scaling parameter to speed things up
    fitDist(protDiffs, funcHypsec, "log10(protein diff in group)", ["muInGroupDiffs", "sigmaInGroupDiffs"], params, plot)
    params['inGroupDiffPrior'] = funcHypsec(params['proteinDiffCandidates'], params['muInGroupDiffs'], params['sigmaInGroupDiffs'])
  
  #fitDist(protGroupDiffs, funcHypsec, "log10(protein diff between groups)", ["muProteinGroupDiffs", "sigmaProteinGroupDiffs"], params, plot)
  
def fitLogitNormal(observedValues, params, plot):
  m = np.mean(observedValues)
  s = np.std(observedValues)
  minBin, maxBin = m - 4*s, m + 4*s
  #minBin, maxBin = -2, 6
  vals, bins = np.histogram(observedValues, bins = np.arange(minBin, maxBin, 0.1), normed = True)
  bins = bins[:-1]
  popt, _ = curve_fit(funcLogitNormal, bins, vals, p0 = (m, s, m - s, s))
  
  resetXICHyperparameters = False
  if popt[0] < popt[2] - 5*popt[3] or popt[0] > popt[2] + 5*popt[3]:
    print("  Warning: muDetect outside of expected region [", popt[2] - 5*popt[3] , ",", popt[2] + 5*popt[3], "]:", popt[0], ".")
    resetXICHyperparameters = True
  
  if popt[1] < 0.1 or popt[1] > 2.0:
    print("  Warning: sigmaDetect outside of expected region [0.1,2.0]:" , popt[1], ".")
    resetXICHyperparameters = True
  
  if resetXICHyperparameters:
    print("    Resetting mu/sigmaDetect hyperparameters to default values of muDetect = muXIC - 1.0 and sigmaDetect = 0.3")
    popt[1] = 0.3
    popt[0] = popt[2] - 1.0
  
  #print("  params[\"muDetectInit\"], params[\"sigmaDetectInit\"] = %f, %f" % (popt[0], popt[1]))
  print("  params[\"muDetect\"], params[\"sigmaDetect\"] = %f, %f" % (popt[0], popt[1]))
  print("  params[\"muXIC\"], params[\"sigmaXIC\"] = %f, %f" % (popt[2], popt[3]))
  #params["muDetectInit"], params["sigmaDetectInit"] = popt[0], popt[1]
  #popt[0], popt[1] = popt[2] - popt[3]*3, popt[3]*1.5
  params["muDetect"], params["sigmaDetect"] = popt[0], popt[1]
  params["muXIC"], params["sigmaXIC"] = popt[2], popt[3]
  if plot:
    poptNormal, _ = curve_fit(funcNorm, bins, vals)
    
    import matplotlib.pyplot as plt
    plt.figure()
    plt.title('Curve fits for muDetect, sigmaDetect, muXIC and sigmaXIC', fontsize = 14)
    plt.bar(bins, vals, width = bins[1] - bins[0], alpha = 0.5, label = 'observed distribution')
    plt.plot(bins, funcLogitNormal(bins, *popt), 'g', label='logit-normal fit', linewidth = 2.0)
    plt.plot(bins, 0.5 + 0.5 * np.tanh((np.array(bins) - popt[0]) / popt[1]), 'm', label = "logit-part fit", linewidth = 2.0)
    plt.plot(bins, funcNorm(bins, popt[2], popt[3]), 'c', label = "normal-part fit", linewidth = 2.0)
    plt.plot(bins, funcNorm(bins, *poptNormal), 'r', label='normal fit (for comparison)', linewidth = 2.0, alpha = 0.5)
    plt.ylabel("relative frequency", fontsize = 14)
    plt.xlabel("log10(intensity)", fontsize = 14)
    plt.legend()
    plt.tight_layout()
    
def fitDist(ys, func, xlabel, varNames, params, plot, x = np.arange(-2,2,0.01)):
  vals, bins = np.histogram(ys, bins = x, normed = True)
  bins = bins[:-1]
  popt, _ = curve_fit(func, bins, vals)
  outputString = ", ".join(["params[\"%s\"]"]*len(popt)) + " = " + ", ".join(["%f"] * len(popt))
  for varName, val in zip(varNames, popt):
    params[varName] = val
  
  if func == funcHypsec:
    fitLabel = "hypsec fit"
  elif func == funcNorm:
    fitLabel = "normal fit"
  elif func == funcGamma:
    fitLabel = "gamma fit"
  else:
    fitLabel = "distribution fit"
  print("  " + outputString % tuple(varNames + list(popt)))
  if plot:    
    import matplotlib.pyplot as plt
    plt.figure()
    plt.title("Curve fit for " + " ".join(varNames), fontsize = 14)
    plt.bar(bins, vals, width = bins[1] - bins[0], label = 'observed distribution')
    plt.plot(bins, func(bins, *popt), 'g', label=fitLabel, linewidth = 2.0)
    if func == funcHypsec:
      poptNormal, _ = curve_fit(funcNorm, bins, vals)
      plt.plot(bins, funcNorm(bins, *poptNormal), 'r', label = 'normal fit (for comparison)', linewidth = 2.0, alpha = 0.5)
      
      if False:
        funcStudentT = lambda x, df, mu, sigma : t.pdf(x, df = df, loc = mu, scale = sigma)
        poptStudentT, _ = curve_fit(funcStudentT, bins, vals)
        print(poptStudentT)
        
        funcCauchy = lambda x, mu, sigma : cauchy.pdf(x, loc = mu, scale = sigma)
        poptCauchy, _ = curve_fit(funcCauchy, bins, vals)
        print(poptCauchy)
        
        plt.plot(bins, funcStudentT(bins, *poptStudentT), 'm', label = 'student-t fit', linewidth = 2.0)
        plt.plot(bins, funcCauchy(bins, *poptCauchy), 'c', label = 'cauchy fit', linewidth = 2.0)
        
        funcLogStudentT = lambda x, df, mu, sigma : t.logpdf(x, df = df, loc = mu, scale = sigma)
        funcLogNorm = lambda x, mu, sigma : norm.logpdf(x, loc = mu, scale = sigma)
        funcLogCauchy = lambda x, mu, sigma : cauchy.logpdf(x, loc = mu, scale = sigma)
        
        plt.ylabel("relative frequency", fontsize = 14)
        plt.xlabel(xlabel, fontsize = 14)
        plt.legend()
        
        plt.figure()
        plt.plot(bins, funcLogHypsec(bins, *popt), 'g', label = 'hypsec log fit', linewidth = 2.0)
        plt.plot(bins, funcLogNorm(bins, *poptNormal), 'r', label = 'normal log fit', linewidth = 2.0)
        plt.plot(bins, funcLogStudentT(bins, *poptStudentT), 'm', label = 'student-t log fit', linewidth = 2.0)
        plt.plot(bins, funcLogCauchy(bins, *poptCauchy), 'c', label = 'cauchy log fit', linewidth = 2.0)
    plt.ylabel("relative frequency", fontsize = 14)
    plt.xlabel(xlabel, fontsize = 14)
    plt.legend()
    plt.tight_layout()

# this is an optimized version of applying parsers.weightedGeomAvg to each of the columns separately
def getProteinQuant(quantMatrixNormalized, quantRows):
  numSamples = len(quantMatrixNormalized[0])
  
  geoAvgQuantRow = np.array([0.0]*numSamples)
  weights = np.array([[1.0 - y.combinedPEP for y in quantRows]] * numSamples).T
  weights[np.isnan(np.array(quantMatrixNormalized))] = np.nan
  
  weightSum = np.nansum(weights, axis = 0)
  weightSum[weightSum == 0] = np.nan
  geoAvgQuantRow = np.exp(np.nansum(np.multiply(np.log(quantMatrixNormalized), weights), axis = 0) / weightSum)
  geoAvgQuantRow = parsers.geoNormalize(geoAvgQuantRow)
  return geoAvgQuantRow
  
def imputeValues(quantMatrixLog, proteinRatios, testProteinRatios):
  logIonizationEfficiencies = quantMatrixLog - np.log10(proteinRatios)
  
  numNonZeros = np.count_nonzero(~np.isnan(logIonizationEfficiencies), axis = 1)[:,np.newaxis] - ~np.isnan(logIonizationEfficiencies)
  np.nan_to_num(logIonizationEfficiencies, False)
  meanLogIonEff = (np.nansum(logIonizationEfficiencies, axis = 1)[:,np.newaxis] - logIonizationEfficiencies) / numNonZeros
  
  logImputedVals = meanLogIonEff + testProteinRatios
  return logImputedVals
    
def funcLogitNormal(x, muLogit, sigmaLogit, muNorm, sigmaNorm):
  return logit(x, muLogit, sigmaLogit) * norm.pdf(x, muNorm, sigmaNorm)

def funcNorm(x, mu, sigma):
  return norm.pdf(x, mu, sigma)
  
def funcHypsec(x, mu, sigma):
  return hypsecant.pdf(x, mu, sigma)
  #return cauchy.pdf(x, mu, sigma)
  #return norm.pdf(x, mu, sigma)

def funcLogHypsec(x, mu, sigma):
  return hypsecant.logpdf(x, mu, sigma)
  #return cauchy.logpdf(x, mu, sigma)
  #return norm.logpdf(x, mu, sigma)
  
def funcGamma(x, shape, sigma):
  return gamma.pdf(x, shape, 0.0, sigma)
  
def logit(x, muLogit, sigmaLogit):
  return 0.5 + 0.5 * np.tanh((np.array(x) - muLogit) / sigmaLogit)




'''
Create Triqler input files by combining Percolator output with the feature 
groups from Quandenser. Requires Triqler and NumPy to be installed.

If Percolator output is unavailable, one can mimick this format by providing
a tab-delimited file with the following columns (including a header row):

PSMId <tab> score <tab> q-value <tab> posterior_error_prob <tab> peptide <tab> proteinIds

The q-value and posterior_error_prob can be set to 0, as they are not used 
here. Furthermore, the file should be sorted by the score column, where higher 
scores indicate more confident hits and the highest score is on top of the list.
'''

from __future__ import print_function

import os
import sys
import numpy as np
import collections

from .. import parsers
from ..triqler import __version__, __copyright__

from . import normalize_intensities as normalize
from . import helpers

def main():
  print('Triqler.convert.quandenser version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  convertQuandenserToTriqler(args.file_list_file, args.in_file, args.psm_files.split(","), args.out_file, params)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  
  requiredNamed = apars.add_argument_group('required arguments')
  
  apars.add_argument('in_file', default=None, metavar = "IN_FILE",
                     help='''Quandenser output file with feature groups.
                          ''')
  
  requiredNamed.add_argument('--file_list_file', metavar='L', 
                     help='Simple text file with spectrum file names in first column and condition in second column.',
                     required = True)
  
  requiredNamed.add_argument('--psm_files', metavar='TARGET,DECOY', 
                     help='Percolator PSM output files, separated commas. Both target and decoy output files are needed, with the target file(s) specified first.',
                     required = True)
  
  apars.add_argument('--out_file', default = "triqler_input.tsv", metavar='OUT', 
                     help='''Path to triqler input file (writing in TSV format).
                          ''')
  
  apars.add_argument('--skip_normalization',
                     help='Skip retention-time based intensity normalization.',
                     action='store_true')
  
  apars.add_argument('--retain_unidentified',
                     help='Keeps features without identification in the Triqler input file.',
                     action='store_true')
                     
  apars.add_argument('--skip_link_pep',
                     help='Skips the linkPEP column from match-between-runs output.',
                     action='store_true')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['simpleOutputFormat'] = args.skip_link_pep
  params['skipNormalization'] = args.skip_normalization
  params['retainUnidentified'] = args.retain_unidentified
  params['plotScatter'] = False
  
  return args, params
  
def convertQuandenserToTriqler(fileListFile, clusterQuantFile, psmsOutputFiles, peptQuantRowFile, params):
  fileInfoList = parsers.parseFileList(fileListFile)
  
  if not params['skipNormalization']:
    clusterQuantFileNormalized = clusterQuantFile.replace(".tsv", ".normalized.tsv")
    if not os.path.isfile(clusterQuantFileNormalized):
      print("Applying retention-time dependent intensity normalization")
      minRunsObservedIn = len(fileInfoList) / 3 + 1
      normalize.normalizeIntensitiesRtimeBased(clusterQuantFile, clusterQuantFileNormalized, minRunsObservedIn, plotScatter = params['plotScatter'])
    else:
      print("Reusing previously generated normalized feature group file:", clusterQuantFileNormalized, ". Remove this file to redo normalization")
    clusterQuantFile = clusterQuantFileNormalized
  else:
    print("Skipping retention-time dependent intensity normalization")
  
  specToPeptideMap = helpers.parsePsmsPoutFiles(psmsOutputFiles)
  
  printTriqlerInputFile(fileInfoList, clusterQuantFile, peptQuantRowFile, specToPeptideMap, params)
  
def parsePeptideLinkPEP(peptLinkPEP):
  spectrumIdx, linkPEP = peptLinkPEP.split(";")
  return int(spectrumIdx), float(linkPEP)

def printTriqlerInputFile(fileInfoList, clusterQuantFile, quantRowFile, specToPeptideMap, params):
  print("Parsing cluster quant file")
  
  writer = parsers.getTsvWriter(quantRowFile)
  if params['simpleOutputFormat']:
    writer.writerow(parsers.TriqlerSimpleInputRowHeaders)
  else:
    writer.writerow(parsers.TriqlerInputRowHeaders)
  
  featureClusterRows = list()
  spectrumToFeatureMatch = dict() # stores the best peptideQuantRow per (peptide, spectrumIdx)-pair
  for featureClusterIdx, featureCluster in enumerate(parsers.parseFeatureClustersFile(clusterQuantFile)):
    if featureClusterIdx % 50000 == 0:
      print("Processing feature group", featureClusterIdx + 1)
    
    rows = list()
    for pc in featureCluster:
      fileIdx = int(pc.fileName)
      for peptLinkPEP in pc.peptLinkPEPs.split(","):
        spectrumIdx, linkPEP = parsePeptideLinkPEP(peptLinkPEP)
        peptide, proteins, searchScore, charge = specToPeptideMap(spectrumIdx)
        if pc.intensity > 0.0 and linkPEP < 1.0 and (params["retainUnidentified"] or peptide != "NA"):
          # run condition charge spectrumId linkPEP featureClusterId search_score intensity peptide proteins
          run, condition, sample, fraction = fileInfoList[fileIdx]
          row = parsers.TriqlerInputRow(sample, condition, charge, spectrumIdx, linkPEP, featureClusterIdx, searchScore, pc.intensity, peptide, proteins)
          rows.append(row)
    
    newRows = list()
    rows = sorted(rows, key = lambda x : (x.run, x.spectrumId, x.linkPEP, -1*x.searchScore, -1*x.intensity))
    prevKey = (-1, -1)
    bestSearchScore = -1e9
    for row in rows:
      if prevKey == (row.run, row.spectrumId):
        if row.searchScore > bestSearchScore:
          bestSearchScore = row.searchScore
          newRows.append(row)
      else:
        newRows.append(row)
        prevKey = (row.run, row.spectrumId)
    
    for row in newRows:
      if params['simpleOutputFormat']:
        writer.writerow(row.toSimpleList())
      else:
        writer.writerow(row.toList())

if __name__ == "__main__":
   main()



from __future__ import print_function

import sys
import signal
import warnings
from multiprocessing import Pool

class MyPool:
  def __init__(self, processes = 1, warningFilter = "default"):
    self.warningFilter = warningFilter
    self.pool = Pool(processes, self.initWorker)
    self.results = []
    
  def applyAsync(self, f, args):
    r = self.pool.apply_async(f, args)
    self.results.append(r)
  
  def initWorker(self):
    return init_worker(self.warningFilter)
  
  def checkPool(self, printProgressEvery = -1):
    try:
      outputs = list()
      for res in self.results:
        outputs.append(res.get(timeout = 1000))
        if printProgressEvery > 0 and len(outputs) % printProgressEvery == 0:
          print(" ", len(outputs),"/", len(self.results), "%.2f" % (float(len(outputs)) / len(self.results) * 100) + "%")
      self.pool.close()
      self.pool.join()
      return outputs
    except (KeyboardInterrupt, SystemExit):
      print("Caught KeyboardInterrupt, terminating workers")
      self.pool.terminate()
      self.pool.join()
      sys.exit()


def init_worker(warningFilter):
  # set warningFilter for the child processes
  warnings.simplefilter(warningFilter)
  
  # causes child processes to ignore SIGINT signal and lets main process handle
  # interrupts instead (https://noswap.com/blog/python-multiprocessing-keyboardinterrupt)
  signal.signal(signal.SIGINT, signal.SIG_IGN)

def addOne(i):
  return i+1

def unitTest():
  pool = MyPool(4)
  for i in range(20):
    pool.applyAsync(addOne, [i])
  results = pool.checkPool()
  print(results)



#!/usr/bin/python

from __future__ import print_function

import itertools

import numpy as np
from scipy.stats import f_oneway, kruskal

from . import parsers
from . import qvality

def doDiffExp(params, peptQuantRows, outputFile, proteinQuantificationMethod, selectComparison, qvalMethod):    
  proteinModifier, getEvalFeatures, evalFunctions = getEvalFunctions(outputFile, params)
  
  proteinOutputRows = proteinQuantificationMethod(peptQuantRows, params, proteinModifier, getEvalFeatures)
  
  if len(params['proteinPosteriorsOutput']) > 0:
    printProteinPosteriors(proteinOutputRows, params)
    
  if len(params['groupPosteriorsOutput']) > 0:
    printGroupPosteriors(proteinOutputRows, params)
  
  if len(params['foldChangePosteriorsOutput']) > 0:
    printFoldChangePosteriors(proteinOutputRows, params)
  
  numGroups = len(params['groups'])
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    if numGroups == 2:
      proteinOutputFile = outputFile
    else:
      proteinOutputFile = getOutputFile(outputFile, groupId1, groupId2)
    
    print("Comparing", params['groupLabels'][groupId1], "to", params['groupLabels'][groupId2])
    print("  output file:", proteinOutputFile)
    
    proteinOutputRowsGroup = selectComparison(proteinOutputRows, (groupId1, groupId2))
    if "trueConcentrationsDict" in params and len(params["trueConcentrationsDict"]) > 0:
      evalFunctions = [lambda protein, evalFeatures : evalTruePositiveTtest(params["trueConcentrationsDict"], protein, groupId1, groupId2, evalFeatures[-2], params)]
    
    printProteinQuantRows(proteinOutputRowsGroup, qvalMethod, evalFunctions, proteinOutputFile, params)

def getOutputFileExtension(outputFile):
  fileName = outputFile.split("/")[-1]
  if "." in fileName:
    return outputFile, "." + fileName.split(".")[-1]
  else:
    return outputFile + ".", "."

def getOutputFile(outputFile, groupId1, groupId2):
  outputFile, outputFileExt = getOutputFileExtension(outputFile)
  return outputFile.replace(outputFileExt, ".%dvs%d%s" % (groupId1 + 1, 
      groupId2 + 1, outputFileExt if len(outputFileExt) > 1 else ""))
  
def getTrueConcentrations(trueConcentrationsDict, protein):
  for key, value in trueConcentrationsDict.items():
    if key in protein:
      return value
  return []
      
def evalTruePositiveANOVA(trueConcentrationsDict, protein):
  trueConcentrations = getTrueConcentrations(trueConcentrationsDict, protein)
  return len(trueConcentrations) > 0

def evalTruePositiveTtest(trueConcentrationsDict, protein, groupId1, groupId2, foldChange, params):
  trueConcentrations = getTrueConcentrations(trueConcentrationsDict, protein)
  if len(trueConcentrations) > 0:
    trueRatios = np.dot(np.matrix(trueConcentrations).T, np.ones_like(np.matrix(trueConcentrations)))
    trueLogRatios = np.log2(trueRatios / trueRatios.T)
    return np.abs(trueLogRatios[groupId1, groupId2]) > params['foldChangeEval'] and \
             trueLogRatios[groupId1, groupId2]*foldChange > 0
  else:
    return False
      
def getEvalFunctions(peptInputFile, params):
  getDEEvalFeatures = lambda quant : [getFoldChange(quant, params), getDiffExp(quant, params)]
  getEvalFeatures = getDEEvalFeatures
  proteinModifier = lambda protein : protein
  
  evalFunctions = []
  return proteinModifier, getEvalFeatures, evalFunctions

def getDiffExp(quants, params):
  quants = parsers.getQuantGroups(quants, params['groups'])
  quantsNotNan = list()
  for q in quants:
    quantsNotNan.append([x for x in q if not np.isnan(x) and not np.isinf(x)])
  
  anovaPvalues = dict()
  anovaPvalues['ANOVA'] = getPval(quantsNotNan)
  
  numGroups = len(params['groups'])
  if numGroups > 2:
    for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
      quantsNotNanFiltered = [quantsNotNan[groupId1], quants[groupId2]]
      anovaPvalues[(groupId1, groupId2)] = getPval(quantsNotNanFiltered)
  return anovaPvalues
  
def getPval(quants):
  anovaFvalue, anovaPvalue = f_oneway(*quants)
  if not np.isnan(anovaPvalue):
    return anovaPvalue
  else:
    #print(quants)
    return 1.0
    
def getFoldChange(quants, params):
  foldChange = dict()
  
  numGroups = len(params['groups'])
  if numGroups > 2:
    maxFoldChange = 0.0
    for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
      foldChange[(groupId1, groupId2)] = getFc(quants, params, groupId1, groupId2)
      maxFoldChange = np.max([maxFoldChange, np.abs(foldChange[(groupId1, groupId2)])])
    foldChange['ANOVA'] = maxFoldChange
  elif numGroups == 2:
    foldChange['ANOVA'] = getFc(quants, params, 0, 1)
  return foldChange

def getFc(quants, params, groupId1, groupId2):
  return np.log2(np.mean([quants[x] for x in params['groups'][groupId1]]) / np.mean([quants[x] for x in params['groups'][groupId2]]))
  
def printProteinQuantRows(proteinOutputRows, qvalMethod, evalFunctions, outputFile, params, qvalThreshold = 0.05):
  writer = parsers.getTsvWriter(outputFile)
  
  evalHeaders = ["log2_fold_change", "diff_exp_prob_" + str(params['foldChangeEval'])]
  if 'pvalues' in qvalMethod:
    evalHeaders[1] = "diff_exp_pval_" + str(params['foldChangeEval'])
    targetPvalues = [x[4] for x in proteinOutputRows]
    reportedQvalsPval, reportedPEPsPval = qvality.getQvaluesFromPvalues(targetPvalues, includePEPs = True)
  
  protOutputHeaders = ["posterior_error_prob", "protein", "num_peptides", "protein_id_posterior_error_prob"] + evalHeaders + parsers.getRunIds(params) + ["peptides"]
  checkCalibration = len(evalFunctions) > 0
  if checkCalibration:
    evalTruePositives = evalFunctions[0]
    writer.writerow(["observed_q_value", "reported_q_value"] + protOutputHeaders)
  else:
    writer.writerow(["q_value"] + protOutputHeaders)
  
  outRows = list()
  observedQvals, reportedQvals, reportedPEPs = list(), list(), list()
  qval, observedQval, sumPEP = 0.0, 0.0, 0.0
  targets, fp, tp = 0, 1, 0
  numTies = 1
  nextScores = [x[0] for x in proteinOutputRows] + [np.nan]
  numSignificant = 0
  for i, (combinedPEP, _, protein, quantRows, evalFeatures, numPeptides, proteinIdPEP, quants, _) in enumerate(proteinOutputRows):
    if 'pvalues_with_fc' in qvalMethod and np.abs(evalFeatures[-2]) < params['foldChangeEval']:
      continue

    if not protein.startswith(params['decoyPattern']):
      sumPEP += combinedPEP
      targets += 1
      qval = sumPEP / targets
      if qval < qvalThreshold:
        numSignificant += 1
      
      if checkCalibration:
        if evalTruePositives(protein, evalFeatures):
          tp += 1
        else:
          fp += 1
        observedQval = float(fp) / (tp+fp)

    if combinedPEP == nextScores[i+1]:
      numTies += 1
    else:
      for _ in range(numTies):
        if 'pvalues' in qvalMethod:
          reportedQvals.append(reportedQvalsPval[i])
        else:
          reportedQvals.append(qval)
        
        if checkCalibration:
          observedQvals.append(observedQval)
      numTies = 1
    
    if 'pvalues' in qvalMethod:
      combinedPEP = reportedPEPsPval[i]
    
    outRows.append(["%.4g" % combinedPEP, protein, numPeptides, "%.4g" % proteinIdPEP] + ["%.4g" % x for x in evalFeatures] + ["%.4g" % x for x in quants] + [x.peptide for x in quantRows])
  
  if checkCalibration:
    observedQvals = qvality.fdrsToQvals(observedQvals)
    for outRow, observedQval, reportedQval in zip(outRows, observedQvals, reportedQvals):
      writer.writerow(["%.4g" % (observedQval), "%.4g" % (reportedQval)] + outRow)
  else:
    for outRow, reportedQval in zip(outRows, reportedQvals):
      writer.writerow(["%.4g" % (reportedQval)] + outRow)
  
  print("  Found", numSignificant, "target proteins as differentially abundant at", str(int(qvalThreshold * 100)) + "% FDR")

def printProteinPosteriors(proteinOutputRows, params):
  print("Writing protein posteriors to", params['proteinPosteriorsOutput'])
  writer = parsers.getTsvWriter(params['proteinPosteriorsOutput'])
  
  writer.writerow(["protein", "group:run"] + ['%.4g' % x for x in params['proteinQuantCandidates']])
  for i, (_, protein, _, _, _, _, _, posteriorDists) in enumerate(proteinOutputRows):
    if posteriorDists:
      pProteinQuantsList, _, _ = posteriorDists
      for run, posterior in zip(parsers.getRunIds(params), pProteinQuantsList):
        writer.writerow([protein, run] + ['%.4g' % p for p in posterior])
    
def printGroupPosteriors(proteinOutputRows, params):
  print("Writing treatment group posteriors to", params['groupPosteriorsOutput'])
  writer = parsers.getTsvWriter(params['groupPosteriorsOutput'])
  
  writer.writerow(["protein", "group"] + ['%.4g' % x for x in params['proteinQuantCandidates']])
  for i, (_, protein, _, _, _, _, _, posteriorDists) in enumerate(proteinOutputRows):
    if posteriorDists:
      _, pProteinGroupQuants, _ = posteriorDists

      numGroups = len(params['groups'])
      for groupId, posterior in zip(range(numGroups), pProteinGroupQuants):
        writer.writerow([protein, params['groupLabels'][groupId]] + ['%.4g' % p for p in posterior])

def printFoldChangePosteriors(proteinOutputRows, params):
  print("Writing fold change posteriors to", params['foldChangePosteriorsOutput'])
  writer = parsers.getTsvWriter(params['foldChangePosteriorsOutput'])
  
  writer.writerow(["protein", "comparison"] + ['%.4g' % x for x in params['proteinDiffCandidates']])
  for i, (_, protein, _, _, _, _, _, posteriorDists) in enumerate(proteinOutputRows):
    if posteriorDists:
      _, _, pProteinGroupDiffs = posteriorDists
      numGroups = len(params['groups'])
      if numGroups >= 2:
        for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
          writer.writerow([protein, params['groupLabels'][groupId1] + "_vs_" + params['groupLabels'][groupId2]] + ['%.4g' % p for p in pProteinGroupDiffs[(groupId1, groupId2)]])




from __future__ import print_function

import sys
import numpy as np

def main(argv):
  np.random.seed(2)
  numPoints = 1001
  
  xs = np.linspace(-5, 5, numPoints)
  probs = generateData(numPoints)
  convProbs = convolveProbs(probs)
  
  print(np.sum(convProbs))
  plt.imshow(convProbs, interpolation = 'none')
  plt.show()  
  
def generateData(numPoints):
  probs = list()
  for i in range(4):
    p1 = np.random.rand(numPoints)
    p1 /= np.sum(p1)
    probs.append(p1)
  return probs

def convolveProbs(probs):
  numPoints = len(probs[0])
  convProbs = np.diag(probs[0])

  for p1 in probs[1:]:
    convProbsCopy = np.copy(convProbs)
    convProbs = np.zeros((numPoints, numPoints))
    rowCumSums = np.zeros((numPoints, numPoints))
    for j in range(numPoints):
      rowCumSums[:j, j] = np.cumsum(convProbsCopy[1:j+1, j][::-1])[::-1]
    for i in range(numPoints):
      convProbs[i, i:] += convProbsCopy[i, i:]*np.cumsum(p1[i:])
      convProbs[i, i:] += rowCumSums[i, i:]*p1[i]
      convProbs[i, i+1:] += np.cumsum(convProbsCopy[i, i:-1])*p1[i+1:]
  return convProbs

if __name__ == "__main__":
  main(sys.argv[1:])



from __future__ import print_function

"""triqler.triqler: provides entry point main()."""

__version__ = "0.4.0"
__copyright__ = '''Copyright (c) 2018-2020 Matthew The. All rights reserved.
Written by Matthew The (matthew.the@scilifelab.se) in the
School of Engineering Sciences in Chemistry, Biotechnology and Health at the 
Royal Institute of Technology in Stockholm.'''

import sys
import os
import collections
import copy
import csv
import multiprocessing
import warnings

import numpy as np

from . import parsers
from . import qvality
from . import hyperparameters
from . import pgm
from . import diff_exp

def main():
  print('Triqler version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  params['warningFilter'] = "ignore"
  with warnings.catch_warnings():
    warnings.simplefilter(params['warningFilter'])
    runTriqler(params, args.in_file, args.out_file)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)

  apars.add_argument('in_file', default=None, metavar = "IN_FILE",
                     help='''List of PSMs with abundances (not log transformed!) 
                             and search engine score. See README for a detailed 
                             description of the columns.
                          ''')
  
  apars.add_argument('--out_file', default = "proteins.tsv", metavar='OUT', 
                     help='''Path to output file (writing in TSV format). 
                             N.B. if more than 2 treatment groups are present, 
                             suffixes will be added before the file extension.
                          ''')
  
  apars.add_argument('--fold_change_eval', type=float, default=1.0, metavar='F',
                     help='log2 fold change evaluation threshold.')
  
  apars.add_argument('--decoy_pattern', default = "decoy_", metavar='P', 
                     help='Prefix for decoy proteins.')
  
  apars.add_argument('--min_samples', type=int, default=2, metavar='N', 
                     help='Minimum number of samples a peptide needed to be quantified in.')
  # Peptides quantified in less than the minimum number will be discarded
  
  apars.add_argument('--num_threads', type=int, default=multiprocessing.cpu_count(), metavar='N', 
                     help='Number of threads, by default this is equal to the number of CPU cores available on the device.')
  
  apars.add_argument('--ttest',
                     help='Use t-test for evaluating differential expression instead of posterior probabilities.',
                     action='store_true')
  
  apars.add_argument('--write_spectrum_quants',
                     help='Write quantifications for consensus spectra. Only works if consensus spectrum index are given in input.',
                     action='store_true')
  
  apars.add_argument('--write_protein_posteriors', default = '', metavar='P_OUT',
                     help='Write raw data of protein posteriors to the specified file in TSV format.')
  
  apars.add_argument('--write_group_posteriors', default = '', metavar='G_OUT',
                     help='Write raw data of treatment group posteriors to the specified file in TSV format.')
  
  apars.add_argument('--write_fold_change_posteriors', default = '', metavar='F_OUT',
                     help='Write raw data of fold change posteriors to the specified file in TSV format.')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['warningFilter'] = "default"
  params['foldChangeEval'] = args.fold_change_eval
  params['t-test'] = args.ttest
  params['minSamples'] = args.min_samples
  params['decoyPattern'] = args.decoy_pattern
  params['numThreads'] = args.num_threads
  params['writeSpectrumQuants'] = args.write_spectrum_quants
  params['proteinPosteriorsOutput'] = args.write_protein_posteriors
  params['groupPosteriorsOutput'] = args.write_group_posteriors
  params['foldChangePosteriorsOutput'] = args.write_fold_change_posteriors
  params['returnPosteriors'] = len(params['proteinPosteriorsOutput']) > 0 or len(params['groupPosteriorsOutput']) > 0 or len(params['foldChangePosteriorsOutput']) > 0
  
  if params['minSamples'] < 2:
    sys.exit("ERROR: --min_samples should be >= 2")
  
  return args, params
  
def runTriqler(params, triqlerInputFile, triqlerOutputFile):  
  from timeit import default_timer as timer

  start = timer()

  if not os.path.isfile(triqlerInputFile):
    sys.exit("Could not locate input file %s. Check if the path is correct." % triqlerInputFile)
  
  params['hasLinkPEPs'] = parsers.hasLinkPEPs(triqlerInputFile)
  if triqlerInputFile.endswith(".pqr.tsv"):
    params['fileList'], params['groups'], params['groupLabels'], peptQuantRows = parsers.parsePeptideQuantFile(triqlerInputFile)
  else:
    peptQuantRowFile = triqlerInputFile + ".pqr.tsv"
    peptQuantRows = convertTriqlerInputToPeptQuantRows(triqlerInputFile, peptQuantRowFile, params)
  
  qvalMethod = 'pvalues' if params['t-test'] else 'avg_pep'
  
  selectComparisonBayesTmp = lambda proteinOutputRows, comparisonKey : selectComparisonBayes(proteinOutputRows, comparisonKey, params['t-test'])
  diff_exp.doDiffExp(params, peptQuantRows, triqlerOutputFile, doPickedProteinQuantification, selectComparisonBayesTmp, qvalMethod = qvalMethod)

  end = timer()
  print("Triqler execution took", end - start, "seconds wall clock time")

def convertTriqlerInputToPeptQuantRows(triqlerInputFile, peptQuantRowFile, params):
  peptQuantRowMap, getPEPFromScore, params['fileList'], params['groupLabels'], params['groups'] = groupTriqlerRowsByFeatureGroup(triqlerInputFile, params['decoyPattern'])
  
  if params['hasLinkPEPs'] and params['writeSpectrumQuants']:
    _, spectrumQuantRows, intensityDiv = _selectBestFeaturesPerRunAndPeptide(
        peptQuantRowMap, getPEPFromScore, params, 
        groupingKey = lambda x : x.spectrumId)
    spectrumQuantRows = _divideIntensities(spectrumQuantRows, intensityDiv)
    spectrumQuantRows = _updateIdentPEPs(spectrumQuantRows, params['decoyPattern'], params['hasLinkPEPs'])
    
    specQuantRowFile = triqlerInputFile + ".sqr.tsv"
    print("Writing spectrum quant rows to file:", specQuantRowFile)
    parsers.printPeptideQuantRows(specQuantRowFile, parsers.getRunIds(params), spectrumQuantRows)
  
  spectrumToFeatureMatch, peptideQuantRows, intensityDiv = _selectBestFeaturesPerRunAndPeptide(peptQuantRowMap, getPEPFromScore, params)
  peptideQuantRows = _selectBestPeptideQuantRowPerFeatureGroup(spectrumToFeatureMatch, peptideQuantRows)
  peptideQuantRows = _divideIntensities(peptideQuantRows, intensityDiv)
  peptideQuantRows = _updateIdentPEPs(peptideQuantRows, params['decoyPattern'], params['hasLinkPEPs'])
  
  print("Writing peptide quant rows to file:", peptQuantRowFile)
  parsers.printPeptideQuantRows(peptQuantRowFile, parsers.getRunIds(params), peptideQuantRows)
  
  return peptideQuantRows

def groupTriqlerRowsByFeatureGroup(triqlerInputFile, decoyPattern):
  print("Parsing triqler input file")
  
  peptQuantRowMap = collections.defaultdict(list)
  seenSpectra = set()
  targetScores, decoyScores = list(), list()
  runCondPairs = list()
  for i, trqRow in enumerate(parsers.parseTriqlerInputFile(triqlerInputFile)):
    if i % 1000000 == 0:
      print("  Reading row", i)
    
    peptQuantRowMap[trqRow.featureClusterId].append(trqRow)
    if (trqRow.run, trqRow.condition) not in runCondPairs:
      runCondPairs.append((trqRow.run, trqRow.condition))
    
    if not np.isnan(trqRow.searchScore) and trqRow.spectrumId not in seenSpectra:
      if _isDecoy(trqRow.proteins, decoyPattern):
        decoyScores.append(trqRow.searchScore)
      else:
        targetScores.append(trqRow.searchScore)
      seenSpectra.add(trqRow.spectrumId)
  
  fileList, groupLabels, groups = _getFilesAndGroups(runCondPairs)
  
  print("Calculating identification PEPs")
  getPEPFromScore = qvality.getPEPFromScoreLambda(targetScores, decoyScores)
    
  return peptQuantRowMap, getPEPFromScore, fileList, groupLabels, groups

def _selectBestFeaturesPerRunAndPeptide(peptQuantRowMap, getPEPFromScore, params, groupingKey = lambda x : x.peptide):
  print("Selecting best feature per run and spectrum")
  numRuns = len(params['fileList'])
  
  minIntensity = 1e100
  noSpectrum = 0
  peptideQuantRows = list()
  spectrumToFeatureMatch = dict() # stores the best peptideQuantRow per (protein, spectrumIdx)-pair
  for featureGroupIdx, trqRows in peptQuantRowMap.items():
    if featureGroupIdx % 100000 == 0:
      print("  featureGroupIdx:", featureGroupIdx)
    
    bestFeaturesPerRun = _selectBestFeaturesPerFeatureGroup(trqRows, 
        getPEPFromScore, params['fileList'], groupingKey, numRuns)
    
    for gKey in bestFeaturesPerRun:
      numRunsPresent = sum(1 for x in bestFeaturesPerRun[gKey] if x[0] < 1.01)
      if numRunsPresent < params['minSamples']:
        continue
      
      pqr = _convertFeatureGroupToPeptideQuantRow(bestFeaturesPerRun[gKey], 
          getPEPFromScore, featureGroupIdx, numRuns)
      
       # some feature clusters might not have a spectrum associated with them
      if pqr.spectrum == 0:
        noSpectrum += 1
        pqr = pqr._replace(spectrum = -100 * noSpectrum)
      
      peptideQuantRows.append(pqr)
      
      minIntensity = min(minIntensity, min([x for x in pqr.quant if x > 0.0]))
      
      # combinedPEP field temporarily contains SVM score
      identPEP = getPEPFromScore(pqr.combinedPEP)
      peptLinkErrorProb = 1.0 - np.prod([1.0 - x for x in pqr.linkPEP if x < 1.01])
      combinedPEP = _combinePEPs(identPEP, peptLinkErrorProb)
      
      # multiple featureGroups can be associated with the same consensus spectrum
      # when two or more analytes match closely in prec m/z and retention time;
      # choose the best featureGroup per (peptide, spectrum)-pair based on combinedPEP
      # note that chimeric spectra can still be associated with multiple peptideQuantRows, 
      # as the protein is included in the key
      key = (",".join(pqr.protein), pqr.spectrum / 100)
      if combinedPEP < spectrumToFeatureMatch.get(key, (-1, -1, 1.01))[2]:
        spectrumToFeatureMatch[key] = (pqr.spectrum, featureGroupIdx, combinedPEP)
  
  # divide intensities by a power of 10 for increased readability of peptide 
  # output file, make sure that the lowest intensity retains two significant 
  # digits after printing with two digits after the decimal point
  intensityDiv = np.power(10, np.floor(np.log10(minIntensity))+1)
  
  return spectrumToFeatureMatch, peptideQuantRows, intensityDiv

def _selectBestFeaturesPerFeatureGroup(trqRows, getPEPFromScore, fileList, groupingKey, numRuns):
  # groupingKey => array([linkPEP, triqlerInputRow])
  bestFeaturesPerRun = collections.defaultdict(lambda : [(1.01, None)]*numRuns)
  
  for trqRow in trqRows:
    fileIdx = fileList.index(trqRow.run)
    gKey = groupingKey(trqRow)      
    bestPEPForRun, bestTrqRowForRun = bestFeaturesPerRun[gKey][fileIdx]

    combinedPEP = _combinePEPs(trqRow.linkPEP, getPEPFromScore(trqRow.searchScore))
    
    samePEPhigherIntensity = (combinedPEP == bestPEPForRun and 
        trqRow.intensity > bestTrqRowForRun.intensity)
    if combinedPEP < bestPEPForRun or samePEPhigherIntensity:
      bestFeaturesPerRun[gKey][fileIdx] = (combinedPEP, trqRow)
  
  return bestFeaturesPerRun

def _convertFeatureGroupToPeptideQuantRow(bestFeaturesPerRun, getPEPFromScore, 
    featureGroupIdx, numRuns):
  intensities, linkPEPs, identPEPs = [0.0]*numRuns, [1.01]*numRuns, [1.01]*numRuns
  first = True
  svmScore = -1e9
  for fileIdx, (_, trqRow) in enumerate(bestFeaturesPerRun):
    if trqRow == None:
      continue
    
    if first:
      charge, proteins, first = trqRow.charge, trqRow.proteins, False
    
    if trqRow.searchScore > svmScore or np.isnan(trqRow.searchScore):
      svmScore, spectrumId, peptide = trqRow.searchScore, trqRow.spectrumId, trqRow.peptide
    
    intensities[fileIdx] = trqRow.intensity
    linkPEPs[fileIdx] = trqRow.linkPEP #_combinePEPs(linkPEP, identPEP)
    identPEPs[fileIdx] = getPEPFromScore(trqRow.searchScore)
  
  # fill in PEPs for missing values
  linkPEPs = _setMissingAsMax(linkPEPs)
  identPEPs = _setMissingAsMax(identPEPs)
  
  return parsers.PeptideQuantRow(svmScore, charge, featureGroupIdx, 
      spectrumId, linkPEPs, intensities, identPEPs, peptide, proteins)
    
def _setMissingAsMax(PEPs):
  maxPEP = max([x for x in PEPs if x <= 1.0])
  return [x if x <= 1.0 else maxPEP for x in PEPs]
  
def _getFilesAndGroups(runCondPairs):
  runCondPairs = sorted(runCondPairs, key = lambda x : (x[1], x[0]))
  fileList = list()
  groupLabels, groups = list(), list()
  for run, cond in runCondPairs:
    if run not in fileList:
      fileList.append(run)
      if cond not in groupLabels:
        groupLabels.append(cond)
        groups.append([])
      groups[groupLabels.index(cond)].append(len(fileList) - 1)
  
  if len(fileList) < 2:
    sys.exit("ERROR: There should be at least two runs.")
  elif len(groups) < 2:
    sys.exit("ERROR: At least two conditions (treatment groups) should be specified.")
  elif min([len(g) for g in groups]) < 2:
    sys.exit("ERROR: Each condition (treatment group) should have at least two runs.")
  
  return fileList, groupLabels, groups

# multiple peptideQuantRows can be associated with the same featureGroup 
# when two or more analytes match closely in prec m/z and retention time;
# choose the best peptideQuantRow per featureGroup based on combinedPEP
def _selectBestPeptideQuantRowPerFeatureGroup(spectrumToFeatureMatch, peptideQuantRows):
  featureGroupToSpectrumIdx = dict()
  for (spectrumIdx, featureGroupIdx, combinedPEP) in spectrumToFeatureMatch.values():
    if combinedPEP < featureGroupToSpectrumIdx.get(featureGroupIdx, (-1, 1.01))[1]:
      featureGroupToSpectrumIdx[featureGroupIdx] = (spectrumIdx, combinedPEP)
  survivingSpectrumIdxs = set([y[0] for x, y in featureGroupToSpectrumIdx.items()])
  #print("Surviving spectrumIdxs:", len(survivingSpectrumIdxs))
  
  peptideQuantRows = filter(lambda x : x.spectrum in survivingSpectrumIdxs, peptideQuantRows)
  
  return peptideQuantRows

def _divideIntensities(peptideQuantRows, intensityDiv = 1e6):
  print("Dividing intensities by %g for increased readability" % intensityDiv)
  newPeptideQuantRows = list()
  for row in peptideQuantRows:
    row = row._replace(quant = list(map(lambda x : x/intensityDiv, row.quant)))
    newPeptideQuantRows.append(row)
  return newPeptideQuantRows

def doPickedProteinQuantification(peptQuantRows, params, proteinModifier, getEvalFeatures):
  notPickedProteinOutputRows = _groupPeptideQuantRowsByProtein(
      peptQuantRows, proteinModifier, params['decoyPattern'])
  
  np.random.shuffle(notPickedProteinOutputRows)
  notPickedProteinOutputRows = sorted(notPickedProteinOutputRows, key = lambda x : x[0], reverse = True)
  
  print("Calculating protein-level identification PEPs")
  pickedProteinOutputRows, proteinPEPs = _pickedProteinStrategy(notPickedProteinOutputRows, params['decoyPattern'])  
  
  print("Fitting hyperparameters")
  hyperparameters.fitPriors(peptQuantRows, params)
  
  print("Calculating protein posteriors")
  posteriors = getPosteriors(pickedProteinOutputRows, proteinPEPs, params)
  
  proteinQuantRows = _updateProteinQuantRows(pickedProteinOutputRows, posteriors, proteinPEPs, getEvalFeatures, params)
  
  return proteinQuantRows

def _updateProteinQuantRows(pickedProteinOutputRows, posteriors, proteinPEPs, 
    getEvalFeatures, params):
  proteinQuantRows = list()
  sumPEP = 0.0
  for (linkPEP, protein, quantRows, numPeptides), (bayesQuantRow, muGroupDiffs, probsBelowFoldChange, posteriorDists), proteinPEP in zip(pickedProteinOutputRows, posteriors, proteinPEPs):
    evalFeatures = getEvalFeatures(bayesQuantRow)
    
    if not params['t-test']:
      evalFeatures[-1] = probsBelowFoldChange
      evalFeatures[-2] = muGroupDiffs
    
    sumPEP += proteinPEP
    if not params['t-test'] or sumPEP / (len(proteinQuantRows) + 1) < 0.05:
      proteinQuantRows.append([linkPEP, protein, quantRows, evalFeatures, numPeptides, proteinPEP, bayesQuantRow, posteriorDists])
    
  proteinQuantRows = sorted(proteinQuantRows, key = lambda x : (x[0], x[1]))
  return proteinQuantRows

def _groupPeptideQuantRowsByProtein(peptQuantRows, proteinModifier, decoyPattern):
  protQuantRows = parsers.filterAndGroupPeptides(peptQuantRows)
  
  proteinRows = list()
  for prot, quantRows in protQuantRows:
    psmIdx = 0
    quantRows = sorted(quantRows, key = lambda x : x.combinedPEP)
    seenPeptides = set()
    usablePeptides = 0
    filteredQuantRows = list()
    for quantRow in quantRows:
      cleanPeptide = parsers.cleanPeptide(quantRow.peptide)
      if cleanPeptide not in seenPeptides:
        seenPeptides.add(cleanPeptide)
        usablePeptides += 1
        filteredQuantRows.append(quantRow)

    if usablePeptides < 1:
      continue
    else:
      quantRows = filteredQuantRows
    protein = proteinModifier(prot)
    numPeptides = usablePeptides

    proteinOutputRow = (_getProteinScore(quantRows), list(quantRows[psmIdx].linkPEP), protein, quantRows, numPeptides)
    proteinRows.append(proteinOutputRow)
  
  return proteinRows

def _getProteinScore(quantRows):
  # logged version performs slightly worse on iPRG2016 set, but might 
  # prevent convergence problems in the case of many peptides for a protein
  return np.log(-1*np.sum([np.log(x.combinedPEP) for x in quantRows]))

def _pickedProteinStrategy(notPickedProteinOutputRows, decoyPattern):
  pickedProteinOutputRows = list()
  targetScores, decoyScores = list(), list()
  seenProteins = set()
  for score, linkPEP, protein, quantRows, numPeptides in notPickedProteinOutputRows:
    evalProtein = protein.replace(decoyPattern, "", 1)
    if evalProtein not in seenProteins:
      seenProteins.add(evalProtein)
      if _isDecoy([protein], decoyPattern):
        decoyScores.append(score)
      else:
        targetScores.append(score)
      pickedProteinOutputRows.append([linkPEP, protein, quantRows, numPeptides])
  
  targetScores = np.array(targetScores)
  decoyScores = np.array(decoyScores)
  _, proteinPEPs = qvality.getQvaluesFromScores(targetScores, decoyScores, includePEPs = True, includeDecoys = True, tdcInput = True)
  
  if len(np.nonzero(proteinPEPs < 1.0)) == 0:
    sys.exit("ERROR: No proteins could be identified with a PEP below 1.0, cannot calculate posteriors.")
  else:
    print("  Identified", qvality.countBelowFDR(proteinPEPs, 0.01), "proteins at 1% FDR")
  
  return pickedProteinOutputRows, proteinPEPs

def getPosteriors(pickedProteinOutputRows, peps, params):
  if params['numThreads'] > 1:
    from . import multiprocessing_pool as pool
    processingPool = pool.MyPool(processes = params['numThreads'], warningFilter = params['warningFilter'])
  
  addDummyPosteriors = 0
  posteriors = list()
  for (linkPEP, protein, quantRows, numPeptides), proteinIdPEP in zip(pickedProteinOutputRows, peps):  
    if proteinIdPEP < 1.0:
      if params['numThreads'] > 1:
        processingPool.applyAsync(pgm.getPosteriors, [quantRows, params])
      else:
        posteriors.append(pgm.getPosteriors(quantRows, params))
        if len(posteriors) % 50 == 0:
          print(" ", len(posteriors),"/", sum(1 for p in peps if p < 1.0), "%.2f" % (float(len(posteriors)) / sum(1 for p in peps if p < 1.0) * 100) + "%")
    else:
      addDummyPosteriors += 1
  
  if params['numThreads'] > 1:
    posteriors = processingPool.checkPool(printProgressEvery = 50)
  
  posteriors.extend([pgm.getDummyPosteriors(params)] * addDummyPosteriors)
  
  return posteriors
  
def selectComparisonBayes(proteinOutputRows, comparisonKey, tTest = False):
  proteinOutputRowsUpdatedPEP = list()
  for (linkPEP, protein, quantRows, evalFeatures, numPeptides, proteinPEP, bayesQuantRow, posteriorDists) in proteinOutputRows:
    evalFeaturesNew = copy.deepcopy(evalFeatures)
    evalFeaturesNew[-1] = evalFeatures[-1][comparisonKey] # probBelowFoldChange
    evalFeaturesNew[-2] = evalFeatures[-2][comparisonKey] # log2_fold_change
    if not tTest:
      combinedPEP = _combinePEPs(evalFeaturesNew[-1], proteinPEP)
    else:
      combinedPEP = evalFeaturesNew[-1]
    
    proteinOutputRowsUpdatedPEP.append([combinedPEP, linkPEP, protein, quantRows, evalFeaturesNew, numPeptides, proteinPEP, bayesQuantRow, posteriorDists])

  proteinOutputRowsUpdatedPEP = sorted(proteinOutputRowsUpdatedPEP, key = lambda x : (x[0], x[1]))
  return proteinOutputRowsUpdatedPEP

# calculate peptide-level identification FDRs and update the linkPEPs with this estimate
def _updateIdentPEPs(peptideQuantRows, decoyPattern, hasLinkPEPs):
  print("Calculating peptide-level identification PEPs")
  
  scoreIdxPairs = list()
  for i, row in enumerate(peptideQuantRows):
    if not np.isnan(row.combinedPEP):
      # row.combinedPEP contains the SVM score
      scoreIdxPairs.append([row.combinedPEP, i, _isDecoy(row.protein, decoyPattern)]) 
  
  scoreIdxPairs = sorted(scoreIdxPairs, reverse = True)
  scoreIdxs = np.argsort([x[1] for x in scoreIdxPairs])
  targetScores = np.array([x[0] for x in scoreIdxPairs if x[2] == False])
  decoyScores = np.array([x[0] for x in scoreIdxPairs if x[2] == True])
  
  _, identPEPs = qvality.getQvaluesFromScores(targetScores, decoyScores, includePEPs = True, includeDecoys = True, tdcInput = True)
  
  print("  Identified", qvality.countBelowFDR(identPEPs, 0.01), "peptides at 1% FDR")
  newPeptideQuantRows = list()
  i = 0
  for row in peptideQuantRows:
    identPEP = 1.0
    if not np.isnan(row.combinedPEP):
      identPEP = identPEPs[scoreIdxs[i]]
      i += 1
    
    if hasLinkPEPs:
      newPeptideQuantRows.append(row._replace(combinedPEP = identPEP)) # using consensus spectra
    else:
      newPeptideQuantRows.append(row._replace(combinedPEP = identPEP, identificationPEP = [_combinePEPs(identPEP, x) for x in row.identificationPEP]))
  return newPeptideQuantRows

def _isDecoy(proteins, decoyPattern):
  isDecoyProt = True
  for protein in proteins:
    if not protein.startswith(decoyPattern):
      isDecoyProt = False
      break
  return isDecoyProt
  
def _combinePEPs(linkPEP, identPEP):
  return 1.0 - (1.0 - linkPEP)*(1.0 - identPEP)

if __name__ == "__main__":
  main()



from __future__ import print_function

import collections

import numpy as np

from .. import parsers
from . import normalize_intensities as normalize
from . import percolator

def parsePsmsPoutFiles(psmsOutputFiles, key = lambda psm : psm.scannr):
  specToPeptideMap = collections.defaultdict(list)
  for psmsOutputFile in psmsOutputFiles:
    for psm in percolator.parsePsmsPout(psmsOutputFile):
      specToPeptideMap[key(psm)] = (psm.peptide, psm.proteins, psm.svm_score, psm.charge)
  return lambda spectrumIdx : specToPeptideMap.get(spectrumIdx, getDefaultPeptideHit())

def getDefaultPeptideHit():
  return ("NA", ["NA"], np.nan, -1) # psm.peptide, psm.PEP, psm.proteins, psm.svm_score, psm.charge
  
def getNormalizationFactorArrays(peptideToFeatureMap, fileInfoList, params):
  _, _, sampleList, fractionList = zip(*fileInfoList)
  rTimeArrays, factorArrays = dict(), dict()
  if not params['skipNormalization']:
    print("Applying retention-time dependent intensity normalization")
    minRunsObservedIn = len(set(sampleList)) / 3 + 1
    for fraction in sorted(list(set(fractionList))):
      factorPairs = normalize.getIntensityFactorPairs(peptideToFeatureMap.values(), sortKey = lambda x : (x[2], x[0].run, -1*x[0].intensity, x[1]), minRunsObservedIn = minRunsObservedIn, fraction = fraction)
      print("Fraction:", fraction, "#runs:", len(factorPairs))
      if params['plotScatter']:
        normalize.plotFactorScatter(factorPairs)
      
      rTimeFactorArrays = normalize.getFactorArrays(factorPairs)
      
      rTimeArrays[fraction], factorArrays[fraction] = dict(), dict()
      for key in rTimeFactorArrays:
        rTimeArrays[fraction][key], factorArrays[fraction][key] = zip(*rTimeFactorArrays[key])
  else:
    print("Skipping retention-time dependent intensity normalization")
  
  return rTimeArrays, factorArrays

def writeTriqlerInputFile(triqlerInputFile, peptideToFeatureMap, rTimeArrays, factorArrays, params):
  writer = parsers.getTsvWriter(triqlerInputFile)
  if params['simpleOutputFormat']:
    writer.writerow(parsers.TriqlerSimpleInputRowHeaders)
  else:
    writer.writerow(parsers.TriqlerInputRowHeaders)
  
  for featureClusterIdx, featureCluster in enumerate(peptideToFeatureMap.values()):
    if featureClusterIdx % 50000 == 0:
      print("Processing feature group", featureClusterIdx + 1)
    newRows = selectBestScorePerRun(featureCluster)
    
    if not params['skipMBR']:
      searchScores = [x[0].searchScore for x in newRows if not np.isnan(x[0].searchScore)]
      if len(searchScores) == 0:
        continue # this can happen if the only PSM has a searchScore <= 0
      worstSearchScore = np.min(searchScores)
    
    for (row, rTime, fraction) in newRows:
      if not params['skipNormalization']:
        newIntensity = normalize.getNormalizedIntensity(rTimeArrays[fraction][row.run], factorArrays[fraction][row.run], rTime, row.intensity)
        row = row._replace(intensity = newIntensity)
      
      if np.isnan(row.searchScore):
        if not params['skipMBR']:
          row = row._replace(searchScore = worstSearchScore)
        else:
          continue
      
      if params['simpleOutputFormat']:
        writer.writerow(row.toSimpleList())
      else:
        writer.writerow(row.toList())

def selectBestScorePerRun(rows):
  newRows = list()
  rows = sorted(rows, key = lambda x : (x[0].run, x[0].spectrumId, x[0].linkPEP, -1*x[0].searchScore))
  prevKey = (-1, -1)
  bestSearchScore = -1e9
  for row in rows:
    if prevKey == (row[0].run, row[0].spectrumId):
      if row[0].searchScore > bestSearchScore:
        bestSearchScore = row[0].searchScore
        newRows.append(row)
    else:
      newRows.append(row)
      prevKey = (row[0].run, row[0].spectrumId)
  return newRows




#!/usr/bin/python

from __future__ import print_function

import itertools

import numpy as np
from scipy.stats import f_oneway, gamma
from scipy.optimize import curve_fit

from . import parsers
from . import convolution_dp
from . import hyperparameters

def getPosteriors(quantRowsOrig, params):
  quantRows, quantMatrix = parsers.getQuantMatrix(quantRowsOrig)
  
  pProteinQuantsList, bayesQuantRow = getPosteriorProteinRatios(quantMatrix, quantRows, params)
  pProteinGroupQuants = getPosteriorProteinGroupRatios(pProteinQuantsList, bayesQuantRow, params)
  pProteinGroupDiffs, muGroupDiffs = getProteinGroupsDiffPosteriors(pProteinGroupQuants, params)
  
  probsBelowFoldChange = getProbBelowFoldChangeDict(pProteinGroupDiffs, params)
  if params['returnPosteriors']:
    posteriorDists = (pProteinQuantsList, pProteinGroupQuants, pProteinGroupDiffs)
  else:
    posteriorDists = None
  
  return bayesQuantRow, muGroupDiffs, probsBelowFoldChange, posteriorDists

def getDummyPosteriors(params):
  bayesQuantRow = [1.0 for g in params['groups'] for x in g]
  numGroups = len(params['groups'])  
  probsBelowFoldChange, muGroupDiffs = dict(), dict()
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    probsBelowFoldChange[(groupId1,groupId2)], muGroupDiffs[(groupId1,groupId2)] = 1.0, 0.0
  posteriorDists = None
  return bayesQuantRow, muGroupDiffs, probsBelowFoldChange, posteriorDists
  
def getPosteriorProteinRatios(quantMatrix, quantRows, params, maxIterations = 50, bayesQuantRow = None):
  numSamples = len(quantMatrix[0])
  bayesQuantRow = np.array([1.0]*numSamples)
  converged = False
  for iteration in range(maxIterations):  
    prevBayesQuantRow = np.copy(bayesQuantRow)
    pProteinQuantsList, bayesQuantRow = getPosteriorProteinRatio(quantMatrix, quantRows, bayesQuantRow, params)
    
    bayesQuantRow = parsers.geoNormalize(bayesQuantRow)
    
    diffInIteration = np.log10(prevBayesQuantRow) - np.log10(bayesQuantRow)
    if np.max(diffInIteration*diffInIteration) < 1e-4:
      converged = True
      #print("Converged after iteration", iteration+1, quantRows[0].protein[0])
      break
  
  if not converged:
    print("Warning: failed to converge for protein", quantRows[0].protein[0])
  
  return pProteinQuantsList, bayesQuantRow

def getPosteriorProteinRatio(quantMatrix, quantRows, geoAvgQuantRow, params):
  numSamples = len(quantMatrix[0])
  
  logGeoAvgs = np.log10([parsers.geomAvg(row) for row in quantMatrix])
  featDiffs = np.log10(quantMatrix) - logGeoAvgs[:,np.newaxis]
  pMissingGeomAvg = pMissing(logGeoAvgs, params["muDetect"], params["sigmaDetect"]) # Pr(f_grn = NaN | t_grn = 1)
  
  pQuantIncorrectId = hyperparameters.funcHypsec(featDiffs, params["muFeatureDiff"], params["sigmaFeatureDiff"]) # Pr(f_grn = x | t_grn = 1)
  #pQuantIncorrectIdOld = hyperparameters.funcLogitNormal(np.log10(quantMatrix), params["muDetect"], params["sigmaDetect"], params["muXIC"], params["sigmaXIC"]) 
  
  xImpsAll = imputeValues(quantMatrix, geoAvgQuantRow, params['proteinQuantCandidates'])
  impDiffs = xImpsAll - np.log10(np.array(quantMatrix))[:,:,np.newaxis]
  pDiffs = hyperparameters.funcHypsec(impDiffs, params["muFeatureDiff"], params["sigmaFeatureDiff"]) # Pr(f_grn = x | m_grn = 0, t_grn = 0)
  
  pProteinQuantsList, bayesQuantRow = list(), list()
  for j in range(numSamples):
    pProteinQuant = params['proteinPrior'].copy() # log likelihood
    for i, row in enumerate(quantMatrix):
      linkPEP = quantRows[i].linkPEP[j]
      identPEP = quantRows[i].identificationPEP[j]
      if identPEP < 1.0:
        pMissings = pMissing(xImpsAll[i,j,:], params["muDetect"], params["sigmaDetect"]) # Pr(f_grn = NaN | m_grn = 1, t_grn = 0)
        if np.isnan(row[j]):
          likelihood = pMissings * (1.0 - identPEP) * (1.0 - linkPEP) + pMissingGeomAvg[i] * (identPEP * (1.0 - linkPEP) + linkPEP)
        else:
          likelihood = (1.0 - pMissings) * pDiffs[i,j,:] * (1.0 - identPEP) * (1.0 - linkPEP) + (1.0 - pMissingGeomAvg[i]) * (pQuantIncorrectId[i][j] * identPEP * (1.0 - linkPEP) + linkPEP)
        
        if np.min(likelihood) == 0.0:
          likelihood += np.nextafter(0,1)
        pProteinQuant += np.log(likelihood)
        
    pProteinQuant -= np.max(pProteinQuant)
    pProteinQuant = np.exp(pProteinQuant) / np.sum(np.exp(pProteinQuant))
    pProteinQuantsList.append(pProteinQuant)
    
    eValue, confRegion = getPosteriorParams(params['proteinQuantCandidates'], pProteinQuant)
    bayesQuantRow.append(eValue)
  
  return pProteinQuantsList, bayesQuantRow

def imputeValues(quantMatrix, proteinRatios, testProteinRatios):
  logIonizationEfficiencies = np.log10(quantMatrix) - np.log10(proteinRatios)
  
  numNonZeros = np.count_nonzero(~np.isnan(logIonizationEfficiencies), axis = 1)[:,np.newaxis] - ~np.isnan(logIonizationEfficiencies)
  np.nan_to_num(logIonizationEfficiencies, False)
  meanLogIonEff = (np.nansum(logIonizationEfficiencies, axis = 1)[:,np.newaxis] - logIonizationEfficiencies) / numNonZeros
  
  logImputedVals = np.tile(meanLogIonEff[:, :, np.newaxis], (1, 1, len(testProteinRatios))) + testProteinRatios
  return logImputedVals

def pMissing(x, muLogit, sigmaLogit):
  return 1.0 - hyperparameters.logit(x, muLogit, sigmaLogit) + np.nextafter(0, 1)

def getPosteriorProteinGroupRatios(pProteinQuantsList, bayesQuantRow, params):
  numGroups = len(params["groups"])
  
  pProteinGroupQuants = list()
  for groupId in range(numGroups):
    filteredProteinQuantsList = np.array([x for j, x in enumerate(pProteinQuantsList) if j in params['groups'][groupId]])
    if "shapeInGroupStdevs" in params:
      pMu = getPosteriorProteinGroupMuMarginalized(filteredProteinQuantsList, params)
    else:
      pMu = getPosteriorProteinGroupMu(params['inGroupDiffPrior'], filteredProteinQuantsList, params)
    pProteinGroupQuants.append(pMu)
  
  return pProteinGroupQuants
  
def getPosteriorProteinGroupMu(pDiffPrior, pProteinQuantsList, params):
  pMus = np.zeros_like(params['proteinQuantCandidates'])
  for pProteinQuants in pProteinQuantsList:
    pMus += np.log(np.convolve(pDiffPrior, pProteinQuants, mode = 'valid'))
  
  #pMus = np.nan_to_num(pMus)
  pMus -= np.max(pMus)
  pMus = np.exp(pMus) / np.sum(np.exp(pMus))
  return pMus

def getPosteriorProteinGroupMuMarginalized(pProteinQuantsList, params):
  pMus = np.zeros((len(params['sigmaCandidates']), len(params['proteinQuantCandidates'])))
  for pProteinQuants in pProteinQuantsList:
    for idx, pDiffPrior in enumerate(params['inGroupDiffPrior']):
      pMus[idx,:] += np.log(np.convolve(pDiffPrior, pProteinQuants, mode = 'valid'))
  
  pSigmas = hyperparameters.funcGamma(params['sigmaCandidates'], params["shapeInGroupStdevs"], params["scaleInGroupStdevs"]) # prior
  pMus = np.log(np.dot(pSigmas, np.exp(pMus)))
  
  pMus -= np.max(pMus)
  pMus = np.exp(pMus) / np.sum(np.exp(pMus))
  
  return pMus
  
def getProteinGroupsDiffPosteriors(pProteinGroupQuants, params):
  numGroups = len(params['groups'])  
  pProteinGroupDiffs, muGroupDiffs = dict(), dict()
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    pDifference = np.convolve(pProteinGroupQuants[groupId1], pProteinGroupQuants[groupId2][::-1])
    pProteinGroupDiffs[(groupId1,groupId2)] = pDifference
    muGroupDiffs[(groupId1,groupId2)], _ = np.log2(getPosteriorParams(params['proteinDiffCandidates'], pDifference) + np.nextafter(0, 1))
  return pProteinGroupDiffs, muGroupDiffs
  
def getProbBelowFoldChangeDict(pProteinGroupDiffs, params):
  probsBelowFoldChange = dict()
  numGroups = len(params["groups"])
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    probsBelowFoldChange[(groupId1, groupId2)] = getPosteriorProteinGroupDiff(pProteinGroupDiffs[(groupId1, groupId2)], params)
  #probsBelowFoldChange['ANOVA'] = getProbBelowFoldChangeANOVA(pProteinGroupQuants, params)
  return probsBelowFoldChange

def getPosteriorProteinGroupDiff(pDifference, params):  
  return sum([y for x, y in zip(params['proteinDiffCandidates'], pDifference) if abs(np.log2(10**x)) < params['foldChangeEval']])

# this is a "pseudo"-ANOVA test which calculates the probability distribution 
# for differences of means between multiple groups. With <=4 groups it seemed
# to return reasonable results, but with 10 groups it called many false positives.
def getProbBelowFoldChangeANOVA(pProteinGroupQuants, params):
  if len(pProteinGroupQuants) > 4:
    print("WARNING: this ANOVA-like test might not behave well if >4 treatment groups are present")
  
  if len(pProteinGroupQuants) >= 2:
    convProbs = convolution_dp.convolveProbs(pProteinGroupQuants)
    bandwidth = np.searchsorted(params['proteinQuantCandidates'], params['proteinQuantCandidates'][0] + np.log10(2**params['foldChangeEval']))
    probBelowFoldChange = 0.0
    for i in range(bandwidth):
      probBelowFoldChange += np.trace(convProbs, offset = i)
  else:
    probBelowFoldChange = 1.0
  return min([1.0, probBelowFoldChange])
  
def getPosteriorParams(proteinQuantCandidates, pProteinQuants):
  return 10**np.sum(proteinQuantCandidates * pProteinQuants), 0.0
  if False:
    eValue, variance = 0.0, 0.0
    for proteinRatio, pq in zip(proteinQuantCandidates, pProteinQuants):
      if pq > 0.001:
        #print(10**proteinRatio, pq)
        eValue += proteinRatio * pq

    for proteinRatio, pq in zip(proteinQuantCandidates, pProteinQuants):
      if pq > 0.001:
        variance += pq * (proteinRatio - eValue)**2
    eValueNew = 10**eValue
    
    return eValueNew, [10**(eValue - np.sqrt(variance)), 10**(eValue + np.sqrt(variance))]




#!/usr/bin/python

from __future__ import print_function

import sys
import os
import warnings
import itertools
import textwrap

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scipy.stats import f_oneway, gamma, norm
from scipy.optimize import curve_fit

from ..triqler import __version__, __copyright__
from .. import parsers
from .. import hyperparameters
from .. import pgm
from .. import diff_exp

def main():
  print('Triqler.distribution.plot_posteriors version %s\n%s' % (__version__, __copyright__))
  print('Issued command:', os.path.basename(__file__) + " " + " ".join(map(str, sys.argv[1:])))
  
  args, params = parseArgs()
  
  params['warningFilter'] = "ignore"
  with warnings.catch_warnings():
    warnings.simplefilter(params['warningFilter'])
    plotPosterior(args.in_file, args.protein_id, params)

def parseArgs():
  import argparse
  apars = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent('''Plots posterior distribution for a given protein <P>. 
      
There are two different options for <IN_FILE>: 
1. use one of the posterior output files of Triqler, generated with the 
   --write_<mode>_posteriors flags. This will only print the particular 
   mode (protein, group or fold change).
2. use a Triqler input file. This estimates hyperparameters from the 
   entire input file and subsequently computes and plots posteriors from 
   all modes.
'''))

  apars.add_argument('in_file', default=None, metavar = "IN_FILE",
                     help='Triqler distribution output file or Triqler input file')
  
  apars.add_argument('--protein_id', metavar='P', 
                     help='Protein ID, can be a partial match, e.g. "P39744" will match "sp|P39744|NOC2_YEAST"',
                     required = True)
  
  apars.add_argument('--fold_change_eval', type=float, default=1.0, metavar='F',
                     help='log2 fold change evaluation threshold')
                     
  apars.add_argument('--decoy_pattern', default = "decoy_", metavar='D', 
                     help='Prefix for decoy proteins (only when Triqler input file is used as input)')
  
  # ------------------------------------------------
  args = apars.parse_args()
  
  params = dict()
  params['returnPosteriors'] = True
  params["foldChangeEval"] = args.fold_change_eval
  params["decoyPattern"] = args.decoy_pattern
  params["trueConcentrationsDict"] = dict()
  params['pMax'] = 0.2 # max probability in violin plots
  params['maxFoldChange'] = 2.0 # max fold change in violin plots
    
  return args, params
  
def plotPosterior(inputFile, protein, params):
  with open(inputFile, 'r') as f:
    line = f.readline()
    headerCols = line.split('\t')
    if len(headerCols) < 2:
      sys.exit("Could not identify input format.")
    elif headerCols[0] == 'run':
      plotPosteriorFromTriqlerInput(inputFile, protein, params)
    elif headerCols[1] == 'group:run':
      params['proteinQuantCandidates'] = np.array(list(map(float, headerCols[2:])))
      plotProteinPosteriors(inputFile, protein, params)
    elif headerCols[1] == 'group':
      params['proteinQuantCandidates'] = np.array(list(map(float, headerCols[2:])))
      plotGroupPosteriors(inputFile, protein, params)
    elif headerCols[1] == 'comparison':
      params['proteinDiffCandidates'] = np.array(list(map(float, headerCols[2:])))
      plotFoldChangePosteriors(inputFile, protein, params)
    else:
      sys.exit("Could not identify input format.")
  
  plt.show()

def plotPosteriorFromTriqlerInput(triqlerInputFile, protein, params):
  if not os.path.isfile(triqlerInputFile):
    sys.exit("Could not locate input file %s. Check if the path to the input file is correct." % triqlerInputFile)
  
  peptQuantRowFile = triqlerInputFile + ".pqr.tsv"
  if not os.path.isfile(peptQuantRowFile):
    sys.exit("Could not locate peptide quantification file %s. Run triqler to generate this file." % peptQuantRowFile)
  
  params["runIds"], params['groups'], params['groupLabels'], peptQuantRows = parsers.parsePeptideQuantFile(peptQuantRowFile)

  matplotlib.rcParams['axes.unicode_minus'] = False
  
  print("Fitting hyperparameters")
  hyperparameters.fitPriors(peptQuantRows, params)
  print("")
  
  peptidePEPThreshold = getPeptidePEPThreshold(peptQuantRows) # needed for the naive method
  
  peptQuantRows = list(filter(lambda x : protein in x.protein[0], peptQuantRows))
  if len(peptQuantRows) == 0:
    sys.exit("Could not find any peptides for the protein")
  
  protQuantRows = parsers.filterAndGroupPeptides(peptQuantRows, lambda x : not x.protein[0].startswith(params['decoyPattern']))
  
  plotPosteriors(protQuantRows, params)
  plotPosteriorCalibration(peptQuantRows, peptidePEPThreshold, params, protein)
  
def getPeptidePEPThreshold(peptQuantRows):
  sumPEP = 0.0
  peptidePEPThreshold = 1.0
  for i, pqr in enumerate(filter(lambda x : not x.protein[0].startswith("decoy_"), sorted(peptQuantRows, key = lambda x : x.combinedPEP))):
    sumPEP += pqr.combinedPEP
    peptidePEPThreshold = pqr.combinedPEP
    if sumPEP / (i+1.0) > 0.05:
      break
  return peptidePEPThreshold

def getNaivePosteriorParams(peptQuantRows, peptidePEPThreshold, params, minQuant = 0.05): 
  quantRows = sorted(filter(lambda x : x.combinedPEP <= peptidePEPThreshold, peptQuantRows), key = lambda x : -1 * np.sum(x.quant))
  seenPeptides = set()
  usablePeptides = 0
  filteredQuantRows = list()
  for quantRow in quantRows:
    if quantRow.peptide not in seenPeptides:
      seenPeptides.add(quantRow.peptide)
      filteredQuantRows.append(quantRow.quant)
  
  #print(len(seenPeptides), "unique peptides")
  
  if len(filteredQuantRows) < 3: # use top3 for protein summarization
    print("Skipping naive quant because < 3 peptides were identified")
    return dict(), dict(), seenPeptides
  
  quant = np.zeros(len(filteredQuantRows[0]))
  for quantRow in filteredQuantRows[:3]: # use top3 for protein summarization
    q = np.array(quantRow)
    q[q == 0] = np.mean(q[q != 0]) # missing value imputation by row mean
    #q[q == 0] = minQuant # missing value imputation by limit of quantification
    quant += q
  quant = parsers.geoNormalize(quant)
  quant = np.log2(quant)
  
  numGroups = len(params["groups"])
  
  naiveRatioMu, naiveRatioSigma = dict(), dict()
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    naiveRatioMu[(groupId1, groupId2)] = np.mean([quant[x] for x in params['groups'][groupId1]]) - np.mean([quant[x] for x in params['groups'][groupId2]])
    naiveRatioSigma[(groupId1, groupId2)] = np.sqrt(np.var([quant[x] for x in params['groups'][groupId1]]) / len(params['groups'][groupId1]) + np.var([quant[x] for x in params['groups'][groupId2]])  / len(params['groups'][groupId2]) )
  
  return naiveRatioMu, naiveRatioSigma, seenPeptides
    
def plotPosteriorCalibration(peptQuantRows, peptidePEPThreshold, params, protein):
  minQuant = np.min(np.array([x.quant for x in peptQuantRows]))
  
  trueConcentrations = diff_exp.getTrueConcentrations(params["trueConcentrationsDict"], protein)
  
  naiveRatioMu, naiveRatioSigma, seenPeptides = getNaivePosteriorParams(peptQuantRows, peptidePEPThreshold, params)
  
  numGroups = len(params["groups"])
  if len(trueConcentrations) == 0:
    trueConcentrations = np.array([1.0]*numGroups)
  
  trueRatios = np.dot(np.matrix(trueConcentrations).T, np.ones_like(np.matrix(trueConcentrations)))
  trueLogRatios = np.log2(trueRatios / trueRatios.T)
  
  plotIdx = 1
  plt.figure(3)
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    realRatio = trueLogRatios[groupId1,groupId2]
    
    plt.subplot(1, ((numGroups - 1)*numGroups)/2, plotIdx)
    
    ymin, ymax = -1*params['pMax'], params['pMax']
    
    x = np.arange(-1*params['maxFoldChange'],params['maxFoldChange'],0.01)
    
    if len(naiveRatioMu) > 0: # use top3 for protein summarization
      naiveDist = -1.0*norm.pdf(x, naiveRatioMu[(groupId1, groupId2)], naiveRatioSigma[(groupId1, groupId2)]) / 100 * np.log2(10)  
      plotViolin(naiveDist, x, label = 'Naive quant', color = 'blue')
    #plt.fill_between(x, naiveDist*0.0, naiveDist, where = np.abs(x) > params['foldChangeEval'], facecolor = 'red', alpha = 0.5)
    
    if len(params["trueConcentrationsDict"]) > 0:
      if np.abs(realRatio) < params['maxFoldChange']:
        plt.plot([ymin, ymax], [realRatio, realRatio], 'k--', label = 'true ratio')
      else:
        numMarkers = 10
        if realRatio < 0:
          sign, marker = -1, matplotlib.markers.CARETDOWN
        else:
          sign, marker = 1, matplotlib.markers.CARETUP
        plt.plot(np.linspace(ymin, ymax, numMarkers), [sign*0.95*params['maxFoldChange']]*numMarkers, 'k--', marker = marker, label = 'true ratio')
      
    
    plt.plot([ymin, ymax], [-1*params['foldChangeEval'], -1*params['foldChangeEval']], color = 'red', alpha = 0.5, label = 'fold change cutoff')
    plt.plot([ymin, ymax], [params['foldChangeEval'], params['foldChangeEval']], color = 'red', alpha = 0.5)
    plt.fill_between([ymin, ymax], [-1*params['foldChangeEval'], -1*params['foldChangeEval']], [params['foldChangeEval'], params['foldChangeEval']], color = 'red', alpha = 0.15)
    
    plt.xlim([ymin, ymax])
    xmin, xmax = plt.gca().get_xlim()
    plt.ylim([-1*params['maxFoldChange'], params['maxFoldChange']])
    
    if plotIdx == 1:
      plt.ylabel("log2(fold change)", fontsize = 14)
      plt.gca().yaxis.set_ticks_position('left')
      for tick in plt.gca().yaxis.get_major_ticks():
          tick.label.set_fontsize(14)
      #plt.gca().spines['right'].set_visible(False)
    else:
      plt.gca().set_yticklabels([])
      plt.gca().yaxis.set_ticks_position('none')
    
    plt.gca().xaxis.set_ticks_position('none') 
    plt.gca().set_xticklabels([])
    
    plotIdx += 1
  
  plt.tight_layout(rect=[0, 0.03, 1, 0.95])
  #plt.subplots_adjust(wspace=0, hspace=0)
  plt.legend(loc = 'lower right', fontsize = 14)
  
def plotPosteriors(protQuantRows, params):
  numGroups = len(params['groups'])
  
  imputedDiffs, observedXICValues = list(), list()
  for prot, quantRows in protQuantRows:
    quantRows, quantMatrix = parsers.getQuantMatrix(quantRows)
    
    print("Protein ID:", prot)
    print("")
    
    print("Peptide absolute abundances")
    printQuantRows(quantMatrix, quantRows)

    quantMatrixNormalized = [parsers.geoNormalize(row) for row in quantMatrix]
    print("Peptide relative abundances")
    printQuantRows(quantMatrixNormalized, quantRows)
    
    #geoAvgQuantRow = getSemiNaiveQuants(quantMatrixNormalized, quantRows)
    
    #print("Protein abundance and p-value semi naive quant")
    #printStats(geoAvgQuantRow, params['groups'])
    
    bayesQuantRow, _, probsBelowFoldChange, posteriorDists = pgm.getPosteriors(quantRows, params)
    pProteinQuantsList, pProteinGroupQuants, pProteinGroupDiffs = posteriorDists
    
    print("Protein abundance (expected value) and p-value")
    printStats(bayesQuantRow, params['groups'])
    
    print("Posterior probability |log2 fold change| < %.2f" % (params['foldChangeEval']))
    if numGroups >= 2:
      for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
        print("  Group %s vs Group %s: %f" % (params['groupLabels'][groupId1], params['groupLabels'][groupId2], probsBelowFoldChange[(groupId1, groupId2)]))
    #print("Probability |log2 fold change| < %.2f:" % (params['foldChangeEval']), probsBelowFoldChange['ANOVA'])
    print("")
    
    plt.figure(1)
    plotPosteriorProteinRatios(pProteinQuantsList, params)
    
    plt.figure(2)
    plotPosteriorProteinGroupsRatios(pProteinGroupQuants, params)
    
    plt.figure(3)
    plotPosteriorProteinGroupsDiffs(pProteinGroupDiffs, params)
    
    #plt.figure(5)
    #q = geoAvgQuantRow
    #q = bayesQuantRow
    #data = [[np.log10(q[x]) for x in group if not np.isnan(q[x])] for group in params['groups']]
    #plt.boxplot(data)

def printQuantRows(quantMatrix, quantRows):
  for i, row in enumerate(quantMatrix):
    print("\t".join(['%.2f' % x for x in quantMatrix[i]]) + '\tcombinedPEP=%.2g\tpeptide=%s' % (quantRows[i].combinedPEP, quantRows[i].peptide))
  print("")
  
def printStats(geoAvgQuantRow, groups):
  print("\t".join(['%.2f' % x for x in geoAvgQuantRow]))
  
  args = parsers.getQuantGroups(geoAvgQuantRow, groups)
  anovaFvalue, anovaPvalue = f_oneway(*args)
  print("p-value:", anovaPvalue)
  print("")
  
  float_formatter = lambda x: "%.2f" % x
  np.set_printoptions(formatter={'float_kind':float_formatter})
  
  geoAvgs = np.matrix([parsers.geomAvg([2**y for y in x]) for x in args])
  ratioMatrix = np.log2(np.transpose(geoAvgs) / geoAvgs)
  #print(ratioMatrix)
  #print("")

def getSemiNaiveQuants(quantMatrixNormalized, quantRows):
  numSamples = len(quantMatrixNormalized[0])
  geoAvgQuantRow = [0]*numSamples
  for i in range(numSamples):
    geoAvgQuantRow[i] = parsers.weightedGeomAvg([x[i] for x in quantMatrixNormalized], [1.0 - y.combinedPEP if not np.isnan(x[i]) else np.nan for x, y in zip(quantMatrixNormalized, quantRows)])
  #geoAvgQuantRow = geoNormalize(geoAvgQuantRow)
  geoAvgQuantRow = np.array(geoAvgQuantRow)
  return geoAvgQuantRow
  
def plotPosteriorProteinGroupsDiffs(pProteinGroupDiffs, params):
  numGroups = len(params['groups'])  
  plt.suptitle("Posteriors for fold change differences between groups", fontsize = 14)
  
  plotIdx = 1
  for groupId1, groupId2 in itertools.combinations(range(numGroups), 2):
    pDifference = pProteinGroupDiffs[(groupId1, groupId2)]
    
    #plt.subplot(numGroups - 1, numGroups - 1, groupId1 * (numGroups - 1) + (groupId2 - 1) + 1)
    plt.subplot(1, ((numGroups - 1)*numGroups)/2, plotIdx)
    plotIdx += 1
    if 'groupLabels' in params:
      title = "%s vs %s" % (params['groupLabels'][groupId1], params['groupLabels'][groupId2])
    else:
      title = "%dvs%d" % (groupId1 + 1, groupId2 + 1)
    plt.title(title, fontsize = 14)
    log2diff = np.log2(np.power(10, params['proteinDiffCandidates']))
    plotViolin(pDifference, log2diff, 'Triqler', 'green')
    #plt.fill_between(log2diff, pDifference*0.0, pDifference, where = np.abs(log2diff) > params['foldChangeEval'], facecolor = 'red', alpha = 0.5)
    
    #if groupId1 + 1 == groupId2: # only print(x label on diagonal plots)
    #  plt.ylabel("log2(protein quant group%d / protein quant group%d)" % (groupId1 + 1, groupId2 + 1))
    plt.ylim([-1*params['maxFoldChange'], params['maxFoldChange']])
    #plt.ylim([np.floor(log2diff[np.argmax(pDifference > 1e-4)]), np.ceil(log2diff[::-1][np.argmax(pDifference[::-1] > 1e-4)])])

def plotViolin(pDifference, log2diff, label, color):
  log2diff = log2diff[np.where(np.abs(pDifference) > 1e-4)]
  pDifference = pDifference[np.where(np.abs(pDifference) > 1e-4)]
  plt.plot(pDifference, log2diff, color = color)
  #plt.fill_between(log2diff, pDifference*0.0, pDifference, alpha = 0.5)
  plt.fill_betweenx(log2diff, pDifference, alpha = 0.5, label = label, color = color)
  
def plotPosteriorProteinGroupsRatios(pProteinGroupQuants, params):
  minProteinRatio, maxProteinRatio = max(params['proteinQuantCandidates']), min(params['proteinQuantCandidates'])
  
  print("Normal distribution fits for posterior distributions of treatment group relative abundances:")
  plt.suptitle("Posteriors for treatment group abundances", fontsize = 14)
  numGroups = len(params['groups'])
  for groupId in range(numGroups):
    plt.subplot(numGroups, 1, groupId + 1)
    plt.title("Group %s" % (params['groupLabels'][groupId]))
    plt.plot(params['proteinQuantCandidates'], pProteinGroupQuants[groupId], label='posterior')
    
    # fit protein group ratios with normal distribution
    popt, pcov = curve_fit(lambda x, mu, sigma : norm.pdf(x, mu, sigma), params['proteinQuantCandidates'], pProteinGroupQuants[groupId] * 100)
    mu, sigma = popt
    
    varNames = ["mu", "sigma"]
    outputString = "  Group %s: " % (params['groupLabels'][groupId]) + ", ".join(["%s"]*len(popt)) + " = " + ", ".join(["%f"] * len(popt))
    print(outputString % tuple(varNames + list(popt)))
    
    plt.plot(params['proteinQuantCandidates'], norm.pdf(params['proteinQuantCandidates'], mu, sigma) / 100, ':', label='normal fit')
    
    minProteinRatio = min([minProteinRatio, params['proteinQuantCandidates'][np.argmax(pProteinGroupQuants[groupId] > 1e-4)]])
    maxProteinRatio = max([maxProteinRatio, params['proteinQuantCandidates'][::-1][np.argmax(pProteinGroupQuants[groupId][::-1] > 1e-4)]])
    
  for groupId in range(numGroups):
    plt.subplot(numGroups, 1, groupId + 1)
    plt.xlim([np.floor(minProteinRatio), np.ceil(maxProteinRatio)])
    if groupId == 0:
      plt.legend()
    if groupId == numGroups - 1:
      plt.xlabel('log10(rel. protein quant)', fontsize = 14)
  plt.tight_layout(rect=[0, 0.03, 1, 0.95])

def plotPosteriorProteinRatios(pProteinQuantsList, params, maxCols = 15):
  minProteinRatio, maxProteinRatio = max(params['proteinQuantCandidates']), min(params['proteinQuantCandidates'])
  cols = min([len(g) for g in params['groups']] + [maxCols])
  rows = len(params['groups'])
  
  plt.suptitle("Posteriors for protein abundances", fontsize = 14)
  for j, pProteinQuants in enumerate(pProteinQuantsList):
    if len([g for g, x in enumerate(params['groups']) if j in x]) > 0:
      groupId = [g for g, x in enumerate(params['groups']) if j in x][0]
    else:
      continue
    
    inGroupIdx = params['groups'][groupId].index(j)
    if inGroupIdx < cols:
      plt.subplot(rows, 1, groupId + 1)
      if inGroupIdx < 5:
        label = params['runIds'][j]
      else:
        label = None
      plt.plot(params['proteinQuantCandidates'], pProteinQuants, label = label)
      minProteinRatio = min([minProteinRatio, params['proteinQuantCandidates'][np.argmax(pProteinQuants > 1e-4)]])
      maxProteinRatio = max([maxProteinRatio, params['proteinQuantCandidates'][::-1][np.argmax(pProteinQuants[::-1] > 1e-4)]])
    
  for groupId in range(rows):
    plt.subplot(rows, 1, groupId + 1)
    plt.title("Group %s" % (params['groupLabels'][groupId]))
    plt.xlim([np.floor(minProteinRatio), np.ceil(maxProteinRatio)])
    plt.legend()
    if groupId == rows - 1:
      plt.xlabel('log10(rel. protein quant)', fontsize = 14)
  plt.tight_layout(rect=[0, 0.03, 1, 0.95])

def plotProteinPosteriors(inputFile, protein, params):
  pProteinQuantsList = list()
  params['groupLabels'] = list()
  params['groups'] = list()
  params['runIds'] = list()
  for fileIdx, (protein, groupRun, posterior) in enumerate(parsers.parsePosteriorFile(inputFile, refProtein = protein)):
    pProteinQuantsList.append(posterior)
    group = ":".join(groupRun.split(":")[:-1])
    run = groupRun.split(":")[-1]
    groupIdx = addGroup(group, params)
    params['groups'][groupIdx].append(fileIdx)
    params['runIds'].append(run)
  
  plotPosteriorProteinRatios(pProteinQuantsList, params)
  
def plotGroupPosteriors(inputFile, protein, params):
  pProteinGroupQuants = list()
  params['groupLabels'] = list()
  params['groups'] = list()
  for protein, group, posterior in parsers.parsePosteriorFile(inputFile, refProtein = protein):
    pProteinGroupQuants.append(posterior)
    addGroup(group, params)
  
  plotPosteriorProteinGroupsRatios(pProteinGroupQuants, params)
  
def plotFoldChangePosteriors(inputFile, protein, params):
  pProteinGroupDiffs = dict()
  params['groupLabels'] = list()
  params['groups'] = list()
  params['maxFoldChange'] = 0.0
  for protein, comparison, posterior in parsers.parsePosteriorFile(inputFile, refProtein = protein):
    group1, group2 = comparison.split("_vs_")
    groupId1 = addGroup(group1, params)
    groupId2 = addGroup(group2, params)
    pProteinGroupDiffs[(groupId1, groupId2)] = posterior
  
    log2diff = np.log2(np.power(10, params['proteinDiffCandidates']))
    params['maxFoldChange'] = max([params['maxFoldChange'], np.abs(np.floor(log2diff[np.argmax(posterior > 1e-4)]))])
    params['maxFoldChange'] = max([params['maxFoldChange'], np.abs(np.ceil(log2diff[::-1][np.argmax(posterior[::-1] > 1e-4)]))])
  
  plotPosteriorProteinGroupsDiffs(pProteinGroupDiffs, params)

def addGroup(group, params):
  if group not in params['groupLabels']:
      params['groupLabels'].append(group)
      params['groups'].append([])
  return params['groupLabels'].index(group)

if __name__ == "__main__":
   main()



#!/usr/bin/python

'''
Some helper functions to parse custom data files
'''

from __future__ import print_function

import sys
import numpy as np
import math
import csv
import os
import itertools
import re
from collections import defaultdict, namedtuple


def getTsvReader(filename):
  # Python 3
  if sys.version_info[0] >= 3:
    return csv.reader(open(filename, 'r', newline = ''), delimiter = '\t')
  # Python 2
  else:
    return csv.reader(open(filename, 'rb'), delimiter = '\t')

def getTsvWriter(filename):
  # Python 3
  if sys.version_info[0] >= 3:
    return csv.writer(open(filename, 'w', newline = ''), delimiter = '\t')
  # Python 2
  else:
    return csv.writer(open(filename, 'wb'), delimiter = '\t')

################################################################################
## input: filename <tab> group <tab> experiment <tab> fraction (one per line) ##
################################################################################

def parseFileList(inputFile):
  reader = getTsvReader(inputFile)
  fileList = list()
  groups = list()
  groupNames = list()
  sampleList = list()
  fractionList = list()
  for fileIdx, row in enumerate(reader):
    fileList.append(os.path.splitext(os.path.basename(row[0]))[0])
    if len(row) >= 3:
      sampleList.append(row[2])
      if len(row) >= 4:
        fractionList.append(row[3])
    
    if row[1] not in groupNames:
      groupNames.append(row[1])
      groups.append([])
    groups[groupNames.index(row[1])].append(fileIdx)
  
  for groupIdx, groupName in enumerate(groupNames):
    groupNames[groupIdx] = str(groupIdx + 1) + ":" + groupName
  
  if len(sampleList) == 0:
    sampleList = fileList
  elif len(sampleList) != len(fileList):
    sys.exit("Sample column was empty for some runs")
  
  if len(fractionList) == 0:
    fractionList = [-1]*len(fileList)
  elif len(fractionList) != len(fileList):
    sys.exit("Fraction column was empty for some runs")
  
  fileInfoList = [[x, getGroupLabel(idx, groups, groupNames), sampleList[idx], fractionList[idx]] for idx, x in enumerate(fileList)]
  return fileInfoList

def getGroupLabel(idx, groups, groupLabels):
  groupIdx = [i for i, g in enumerate(groups) if idx in g][0]
  return groupLabels[groupIdx]

def getQuantGroups(quantRow, groups, transform = np.log2):
  args = list()
  for group in groups:
    args.append([transform(quantRow[i]) for i in group if not np.isnan(quantRow[i])])
  return args

def getRunIds(params):
  return [getGroupLabel(idx, params['groups'], params['groupLabels']) + ":" + x.split("/")[-1] for idx, x in enumerate(params['fileList'])]

############################
## Feature cluster files  ##
############################

PrecursorCandidate = namedtuple("PrecursorCandidate", "fileIdx fileName precMz charge rTime intensity peptLinkPEPs")

def parseFeatureClustersFileHandle(reader):
  rows = list()
  name = ""
  for row in reader:
    if len(row) > 5:
      if float(row[4]) > 0.0:
        rows.append(PrecursorCandidate(-1, row[0], float(row[1]), int(row[2]), float(row[3]), float(row[4]), row[5]))
    elif len(row) == 1:
      name = row[0]
    else:
      if len(rows) > 0:
        yield rows
      rows = list()

def parseFeatureClustersFile(clusterQuantFile):
  reader = getTsvReader(clusterQuantFile)
  i = 0
  for x in parseFeatureClustersFileHandle(reader):
    #if i > 10000:
    #  break
    i += 1
    yield x

##############################################
## Dinosaur scan to precursor mapping file  ##
##############################################

MappedPrecursor = namedtuple("MappedPrecursor", "scanNr precMz charge rTime intensity")

def getMappedPrecursorOriginalFile(mappedPrecursorFile):
  reader = getTsvReader(mappedPrecursorFile)
  fileNameLine = next(reader)
  return os.path.splitext(fileNameLine[0].split("=")[-1])[0]

# specId, fmz, fz, frt, fint
def parseMappedPrecursorFile(mappedPrecursorFile):
  reader = getTsvReader(mappedPrecursorFile)
  next(reader) # filename
  next(reader) # headers
  for row in reader:
    if " scan=" in row[0]:
      row[0] = row[0].split(" scan=")[1].split()[0]
    yield MappedPrecursor(int(row[0]), float(row[1]), int(row[2]), float(row[3]), float(row[4]))
    
##########################
## Triqler input files  ##
##########################

TriqlerSimpleInputRowHeaders = "run condition charge searchScore intensity peptide proteins".split(" ")
TriqlerInputRowHeaders = "run condition charge spectrumId linkPEP featureClusterId searchScore intensity peptide proteins".split(" ")
TriqlerInputRowBase = namedtuple("TriqlerInputRow", TriqlerInputRowHeaders)

class TriqlerInputRow(TriqlerInputRowBase):
  def toList(self):
    l = list(self)
    return l[:-1] + l[-1]
  
  def toSimpleList(self):
    l = list(self)
    return l[:3] + l[6:-1] + l[-1]
  
  def toString(self):
    return "\t".join(map(str, self.toList()))

def parseTriqlerInputFile(triqlerInputFile):
  reader = getTsvReader(triqlerInputFile)
  headers = next(reader)
  hasLinkPEPs = "linkPEP" in headers
  getUniqueProteins = lambda x : list(set([p for p in x if len(p.strip()) > 0]))
  intensityCol = 7 if hasLinkPEPs else 4
  seenPeptChargePairs = dict()
  for i, row in enumerate(reader):
    intensity = float(row[intensityCol])
    if intensity > 0.0:
      if hasLinkPEPs:
        proteins = getUniqueProteins(row[9:])
        yield TriqlerInputRow(row[0], row[1], int(row[2]), int(row[3]), float(row[4]), int(row[5]), float(row[6]), intensity, row[8], proteins)
      else:
        key = (int(row[2]), row[5])
        if key not in seenPeptChargePairs:
          seenPeptChargePairs[key] = len(seenPeptChargePairs)
        proteins = getUniqueProteins(row[6:])
        yield TriqlerInputRow(row[0], row[1], int(row[2]), (i+1) * 100, 0.0, seenPeptChargePairs[key], float(row[3]), intensity, row[5], proteins)

def hasLinkPEPs(triqlerInputFile):
  reader = getTsvReader(triqlerInputFile)
  headers = next(reader)
  return "linkPEP" in headers

#####################################
## Triqler posterior output files  ##
#####################################

def parsePosteriorFile(peptideQuantFile, refProtein = ""):
  reader = getTsvReader(peptideQuantFile)
  headers = next(reader)
  
  for row in reader:
    protein = row[0]
    if len(refProtein) == 0 or refProtein in protein:
      yield protein, row[1], np.array(list(map(float, row[2:])))

##############################
## Peptide quant row files  ##
##############################

PeptideQuantRowHeaders = "combinedPEP charge featureGroup spectrum linkPEP quant identificationPEP peptide protein".split(" ")
PeptideQuantRowBase = namedtuple("PeptideQuantRow", PeptideQuantRowHeaders)

class PeptideQuantRow(PeptideQuantRowBase):
  def toList(self):
    l = list(self)
    return l[:4] + list(map(lambda x : '%.5g' % x, l[4])) + list(map(lambda x : '%.2f' % x, l[5])) + list(map(lambda x : '%.5g' % x, l[6])) + l[7:8] + [";".join(map(lambda x : x.replace(";", "_"), l[8]))]

  def toString(self):
    return "\t".join(map(str, self.toList()))

def getPeptideQuantRowHeaders(runs):
  return PeptideQuantRowHeaders[:4] + runs + runs + runs + PeptideQuantRowHeaders[7:]

def parsePeptideQuantFile(peptideQuantFile):
  reader = getTsvReader(peptideQuantFile)
  header = next(reader)
  numRuns = int((header.index("peptide") - header.index("spectrum") - 1) / 3)
  peptideQuantRows = list()
  for row in reader:
    if len(row) > 6+3*numRuns:
      proteins = row[5+3*numRuns:]
    else:
      proteins = row[5+3*numRuns].split(";")
    peptideQuantRows.append(PeptideQuantRow(float(row[0]), int(row[1]), int(row[2]), int(row[3]), np.array(list(map(float, row[4:4+numRuns]))), np.array(list(map(float, row[4+numRuns:4+2*numRuns]))), np.array(list(map(float, row[4+2*numRuns:4+3*numRuns]))), row[4+3*numRuns], proteins))

  runIdsWithGroup = header[4:4+numRuns]
  maxGroups = len(set([runId.split(":")[0] for runId in runIdsWithGroup]))
  runIds = list()
  groups, groupLabels = [None] * maxGroups, [None] * maxGroups
  groupNames = list()
  for fileIdx, runId in enumerate(runIdsWithGroup):
    runIdSplitted = runId.split(":")
    if len(runIdSplitted) == 3:
      groupIdx, groupName, runId = runIdSplitted
      groupIdx = int(groupIdx) - 1
    else:
      groupName, runId = runIdSplitted
      if groupName not in groupNames:
        groupNames.append(groupName)
      groupIdx = groupNames.index(groupName)
    runIds.append(runId)
    if groupName not in groupLabels:
      groupLabels[groupIdx] = groupName
      groups[groupIdx] = []
    groups[groupIdx].append(fileIdx)
  return runIds, groups, groupLabels, peptideQuantRows

def getPeptideQuantFileHeaders(peptideQuantFile):
  reader = getTsvReader(peptideQuantFile)
  header = next(reader)
  return header

def printPeptideQuantRows(peptOutputFile, headers, peptideQuantRows):
  writer = getTsvWriter(peptOutputFile)
  writer.writerow(getPeptideQuantRowHeaders(headers))
  for row in peptideQuantRows:
    writer.writerow(row.toList())
    
def filterAndGroupPeptides(peptQuantRows, peptFilter = lambda x : True):
  validPqr = lambda x : len(x.protein) == 1 and x.protein[0] != "NA" and x.combinedPEP < 1.0
  peptQuantRows = filter(lambda x : validPqr(x) and peptFilter(x), peptQuantRows)
  protQuantRows = itertools.groupby(sorted(peptQuantRows, key = lambda x : x.protein[0]), key = lambda x : x.protein[0])
  return protQuantRows

###################################
## Quant matrix helper functions ##
###################################

def getQuantMatrix(quantRows, condenseChargeStates = True, retainBestChargeState = True):
  quantRows = list(quantRows) # contains full information about the quantification row
  if condenseChargeStates:
    peptQuantRowGroups = itertools.groupby(sorted(quantRows, key = lambda x : cleanPeptide(x.peptide)), key = lambda x : cleanPeptide(x.peptide))
    quantRows = list() # contains full information about the quantification row
    quantMatrix = list() # contains just the quantification values
    for _, pqrs in peptQuantRowGroups:
      pqrs = list(pqrs)
      if retainBestChargeState:
        pqrs = sorted(pqrs, key = lambda x : x.combinedPEP)
        quantMatrix.append([x if x > 0.0 else np.nan for x in pqrs[0].quant])
        quantRows.append(pqrs[0])
      else:
        quantMatrix.append([x if x > 0.0 else np.nan for x in pqrs[0].quant])
        if len(pqrs) > 1:
          peptQuantMatrix = [x.quant for x in pqrs]
          geomAvgQuant = list()
          for i in range(len(quantMatrix[-1])):
            quantsRun = [x[i] if x[i] > 0.0 else np.nan for x in peptQuantMatrix]
            if len(quantsRun) > 0:
              geomAvgQuant.append(weightedGeomAvg(quantsRun, [x.combinedPEP for x in pqrs]))
            else:
              geomAvgQuant.append(0.0)
          quantMatrix[-1] = geomAvgQuant
        quantRows.append(pqrs[0]._replace(qval = geomAvg([x.combinedPEP for x in pqrs])))
  else:
    quantMatrix = [[y if y > 0.0 else np.nan for y in x.quant] for x in quantRows]
  
  quantRows, quantMatrix = zip(*[(x, np.array(y)) for x, y in sorted(zip(quantRows, quantMatrix), key = lambda x : x[0].combinedPEP)])
  quantRows = list(quantRows)
  quantMatrix = list(quantMatrix)
  return quantRows, quantMatrix

def weightedGeomAvg(row, weights):
  weightSum = np.nansum(weights)
  if weightSum > 0:
    return np.exp(np.nansum(np.log(row) * weights) / weightSum)
  else:
    return np.nan

def geomAvg(row):
  return np.exp(np.nanmean(np.log(row)))

def geoNormalize(row):
  return row / geomAvg(row)

def cleanPeptide(peptide):
  if peptide[1] == "." and peptide[-2] == ".":
    peptide = peptide[2:-2]
  return re.sub('\[[-0-9]*\]', '', peptide)



# This is a python port of the C++ code of qvality (https://github.com/percolator/percolator/)
# It does not include the mix-max corrections, nor the pi0 corrections

from __future__ import print_function

import subprocess
import tempfile
import csv
import os
import sys

import numpy as np
import bisect

tao = 2.0 / (1 + np.sqrt(5.0)) # inverse of golden section
scaleAlpha = 1
stepEpsilon = 1e-8
gRange = 35.0
weightSlope = 1e1

VERB = 3

# pi0 estimation parameters
numLambda = 100
maxLambda = 0.5

# this function returns PEPs in ascending order (lowest PEP first)
def getQvaluesFromScores(targetScores, decoyScores, includePEPs = False, includeDecoys = False, tdcInput = False, pi0 = 1.0, plotRegressionCurve = False, numBins = 500):
  if type(targetScores) is not np.ndarray:
    targetScores = np.array(targetScores)
  if type(decoyScores) is not np.ndarray:
    decoyScores = np.array(decoyScores)
  
  if len(targetScores) == 0:
    sys.exit("ERROR: no target hits available for PEP calculation")
  
  if len(decoyScores) == 0:
    sys.exit("ERROR: no decoy hits available for PEP calculation")
  
  targetScores.sort()
  decoyScores.sort()
  allScores = np.concatenate((targetScores, decoyScores))
  allScores.sort()
  
  medians, negatives, sizes = binData(allScores, decoyScores, numBins)
  medians, negatives, sizes = np.array(medians), np.array(negatives), np.array(sizes)
  
  # sort in descending order, highest score first
  if includeDecoys:
    evalScores = allScores[::-1]
    #evalScores = np.array([x[0] for x in combined])
  else:
    evalScores = targetScores[::-1]
  
  if VERB > 3:
    print(medians, negatives, sizes)
  
  variables = roughnessPenaltyIRLS(medians, negatives, sizes)
  
  if pi0 < 1.0:
    factor = pi0 * float(len(targetScores)) / len(decoyScores)
  else:
    factor = 1.0
  
  if plotRegressionCurve:
    scoresForPlot = evalScores.copy()
    
  probs = factor * np.exp(splineEval(evalScores, medians, variables))
  probs = monotonize(probs)
  
  if plotRegressionCurve:
    import matplotlib.pyplot as plt
    plt.plot(medians, (1.0*negatives) / sizes, '*-')
    plt.plot(scoresForPlot, probs)
    
    plt.figure()
    plt.plot(medians, (1.0*negatives) / sizes, '*-')
    plt.plot(scoresForPlot, probs)
    plt.yscale("log")
    plt.show()
  return None, probs
  
def getQvaluesFromPvalues(pvalues, includePEPs = False):
  targetScores = sorted(pvalues)
  pi0 = estimatePi0(targetScores)
  if VERB > 2:
    print("Estimating pi0 = %f" % pi0)
  
  step = 1.0 / 2.0 / len(pvalues)
  decoyScores = np.arange(step, 1 - step + 1e-10, step*2)

  targetScores = pvaluesToScores(targetScores)
  decoyScores = pvaluesToScores(decoyScores)
  return getQvaluesFromScores(targetScores, decoyScores, includePEPs, includeDecoys = False, pi0 = pi0)

def pvaluesToScores(pvalues):
  return np.array(list(map(lambda x : -1*np.log(x / (1 - x)), pvalues)))
  
def monotonize(peps):
  return np.minimum(1.0, np.maximum.accumulate(peps))

def binData(allScores, decoyScores, numBins):
  binEdges = list(map(lambda x : int(np.floor(x)), np.linspace(0, len(allScores), numBins+1)))
  bins = list()
  startIdx = 0
  for endIdx in binEdges[1:]:
    if startIdx < endIdx:
      while endIdx < len(allScores) and allScores[endIdx-1] == allScores[endIdx]:
        endIdx += 1
      bins.append(allScores[startIdx:endIdx])
      startIdx = endIdx
  
  results = list()
  for b in bins:
    m = np.median(b)
    numNegs = np.searchsorted(decoyScores, b[-1], side = 'right') - np.searchsorted(decoyScores, b[0], side = 'left')
    numTot = len(b)
    results.append([m, numNegs, numTot])
  return zip(*results)

def roughnessPenaltyIRLS(medians, negatives, sizes):
  Q, R = initQR(medians)
  g, w, z, gamma, p, gnew = initg(negatives, sizes)
  variables = (Q, R, g, w, z, gamma, p, gnew)
  
  p1 = 1.0 - tao
  p2 = tao
  
  cv1 = evaluateSlope(medians, negatives, sizes, variables, -scaleAlpha * np.log(p1))
  cv2 = evaluateSlope(medians, negatives, sizes, variables, -scaleAlpha * np.log(p2))
  
  alpha = alphaLinearSearchBA(0.0, 1.0, p1, p2, cv1, cv2, medians, negatives, sizes, variables)
  if VERB > 3:
    print("Alpha selected to be", alpha)
  variables = iterativeReweightedLeastSquares(medians, negatives, sizes, variables, alpha)
  return variables

def alphaLinearSearchBA(min_p, max_p, p1, p2, cv1, cv2, medians, negatives, sizes, variables):
  # Minimize Slope score
  # Use neg log of 0<p<1 so that we allow for searches 0<alpha<inf
  oldCV = 0.0
  if cv2 < cv1:
    # keep point 2
    min_p = p1
    p1 = p2
    p2 = min_p + tao * (max_p - min_p)
    oldCV = cv1
    cv1 = cv2
    cv2 = evaluateSlope(medians, negatives, sizes, variables, -1*scaleAlpha*np.log(p2))
    if VERB > 3:
      print("New point with alpha=", -scaleAlpha*np.log(p2), ", giving slopeScore=", cv2)
  else:
    # keep point 1
    max_p = p2
    p2 = p1
    p1 = min_p + (1 - tao) * (max_p - min_p)
    oldCV = cv2
    cv2 = cv1
    cv1 = evaluateSlope(medians, negatives, sizes, variables, -1*scaleAlpha*np.log(p1))
    if VERB > 3:
      print("New point with alpha=", -scaleAlpha*np.log(p1), ", giving slopeScore=", cv1)
  if (oldCV - min(cv1, cv2)) / oldCV < 1e-5 or abs(p2 - p1) < 1e-10:
    return -scaleAlpha*np.log(p1) if cv1 < cv2 else -scaleAlpha*np.log(p2)
  return alphaLinearSearchBA(min_p, max_p, p1, p2, cv1, cv2, medians, negatives, sizes, variables)

def evaluateSlope(medians, negatives, sizes, variables, alpha):
  # Calculate a spline for current alpha
  variables = iterativeReweightedLeastSquares(medians, negatives, sizes, variables, alpha)
  
  _, _, g, _, _, _, _, _ = variables
  
  # Find highest point (we only want to evaluate things to the right of that point)
  n = len(medians)
  mixg = 1 # Ignore 0 and n-1
  maxg = g[mixg]
  for ix in range(mixg, n-1):
    if g[ix] >= maxg:
      maxg = g[ix]
      mixg = ix
  maxSlope = -10e6
  slopeix = -1
  for ix in range(mixg+1, n-2):
    slope = g[ix-1]-g[ix]
    if slope > maxSlope:
      maxSlope = slope
      slopeix = ix
  
  # Now score the fit based on a linear combination between
  # The bump area and alpha
  if VERB > 3:
    print("mixg=", mixg, ", maxg=", maxg, ", maxBA=", maxSlope, " at ix=", slopeix, ", alpha=", alpha)
  
  return maxSlope * weightSlope + alpha

def iterativeReweightedLeastSquares(medians, negatives, sizes, variables, alpha, epsilon = stepEpsilon, maxiter = 50):
  Q, R, g, w, z, gamma, p, gnew = variables
  for it in range(maxiter):
    g = gnew
    p, z, w = calcPZW(g, negatives, sizes)
    aWiQ = np.multiply((alpha / w)[:,None], Q)
    M = R + Q.T.dot(aWiQ)
    gamma = np.linalg.solve(M, Q.T.dot(z))
    gnew = z - aWiQ.dot(gamma)
    gnew = np.minimum(gRange, np.maximum(-1*gRange, gnew))
    difference = g - gnew
    step = np.linalg.norm(difference) / len(medians)
    if VERB > 3:
      print("Step size:", step)
    if step < epsilon:
      return (Q, R, g, w, z, gamma, p, gnew)
  
  if VERB > 1:
    print("Warning: IRLS did not converge with maxIter =", maxiter)
  
  return (Q, R, g, w, z, gamma, p, gnew)

def calcPZW(g, negatives, sizes, epsilon = 1e-15):
  e = np.exp(g)
  p = np.minimum(1 - epsilon, np.maximum(epsilon, e / (1+e)))
  w = np.maximum(epsilon, sizes * p * (1 - p))
  z = np.minimum(gRange, np.maximum(-1*gRange, g + (negatives - p * sizes) / w))
  return p, z, w

def initg(negatives, sizes):
  n = len(negatives)
  g = np.zeros(n)
  w = np.zeros(n)
  z = np.ones(n) * 0.5
  gamma = np.zeros(n-2)
  
  p = (negatives + 0.05) / (sizes + 0.1)
  gnew = np.log(p / (1-p))
  return g, w, z, gamma, p, gnew
  
def initQR(medians):
  n = len(medians)
  dx = medians[1:] - medians[:-1]
  Q = np.zeros((n, n -2))
  Q[range(n-2), range(n-2)] = 1.0 / dx[:-1]
  Q[range(1,n-1), range(n-2)] = - 1.0 / dx[:-1] - 1.0 / dx[1:]
  Q[range(2,n), range(n-2)] = 1.0 / dx[1:]
  
  R = np.zeros((n-2, n-2))
  R[range(n-2), range(n-2)] = (dx[:-1] + dx[1:]) / 3
  R[range(n-3), range(1,n-2)] = dx[1:-1] / 6
  R[range(1,n-2), range(n-3)] = dx[1:-1] / 6
  return Q, R

def splineEval(scores, medians, variables):
  _, _, g, _, _, gamma, _, _ = variables
  #score = np.exp(score)
  n = len(medians)
  rights = np.searchsorted(medians, scores)
  
  derr = (g[1] - g[0]) / (medians[1] - medians[0]) - (medians[1] - medians[0]) / 6 * gamma[0]
  scores[rights == 0] = g[0] - (medians[0] - scores[rights == 0]) * derr # reuse "scores" array to save memory
  
  derl = (g[-1] - g[-2]) / (medians[-1] - medians[-2]) + (medians[-1] - medians[-2]) / 6 * gamma[-3]
  scores[rights == n] = g[-1] + (scores[rights == n] - medians[-1]) * derl
  
  idxs = np.where((rights > 0) & (rights < n))
  rights = rights[idxs] # reuse "rights" array to save memory
  hs = medians[rights] - medians[rights - 1]
  
  drs = medians[rights] - scores[idxs]
  dls = scores[idxs] - medians[rights - 1]
  
  gamr = np.zeros_like(hs)
  gamr[rights < (n - 1)] = gamma[rights[rights < (n - 1)] - 1]
  
  gaml = np.zeros_like(hs)
  gaml[rights > 1] = gamma[rights[rights > 1] - 2]
  
  scores[idxs] = (dls * g[rights] + drs * g[rights - 1]) / hs - dls * drs / 6 * ((1.0 + dls / hs) * gamr + (1.0 + drs / hs) * gaml)
  
  return scores

def estimatePi0(pvalues, numBoot = 100):
  pvalues = np.array(pvalues)
  lambdas, pi0s = list(), list()
  numPvals = len(pvalues)
  
  for lambdaIdx in range(numLambda + 1):
    l = ((lambdaIdx + 1.0) /  numLambda) * maxLambda
    startIdx = np.searchsorted(pvalues, l)
    Wl = numPvals - startIdx
    pi0 = Wl / (1.0 - l) / numPvals
    if pi0 > 0.0:
      lambdas.append(l)
      pi0s.append(pi0)
  
  if len(pi0s) == 0:
    print("Error in the input data: too good separation between target and decoy PSMs.\nImpossible to estimate pi0, setting pi0 = 1")
    return 1.0
  
  minPi0 = min(pi0s)
  
  mse = [0.0] * len(pi0s)
  # Examine which lambda level is most stable under bootstrap
  for boot in range(numBoot):
    pBoot = bootstrap(pvalues)
    n = len(pBoot)
    for idx, l in enumerate(lambdas):
      startIdx = np.searchsorted(pvalues, l)
      Wl = numPvals - startIdx
      pi0Boot = Wl / n / (1 - l)
      # Estimated mean-squared error.
      mse[idx] += (pi0Boot - minPi0) * (pi0Boot - minPi0)
  return max([min([pi0s[np.argmin(mse)], 1.0]), 0.0])

def bootstrap(allVals, maxSize = 1000):
  return sorted(np.random.choice(allVals, min([len(allVals), maxSize]), replace = False))
  
def getQvaluesFromPvaluesQvality(pvalues, includePEPs = False):
  fdp, pvalFile = tempfile.mkstemp()
  with open(pvalFile, 'w') as w:
    for pval in pvalues:
      w.write(str(pval) + '\n')
  
  fdq, qvalFile = tempfile.mkstemp()
  rc = subprocess.call("qvality %s > %s" % (pvalFile, qvalFile), shell=True)
  if includePEPs:
    qvals, peps = parseQvalues(qvalFile, includePEPs = includePEPs)
  else:
    qvals = parseQvalues(qvalFile)
  
  os.unlink(pvalFile)
  os.unlink(qvalFile)
  if includePEPs:
    return qvals, peps
  else:
    return qvals
    
def getQvaluesFromScoresQvality(targetScores, decoyScores, includePEPs = False, includeDecoys = False, tdcInput = False):
  fdp, targetFile = tempfile.mkstemp()
  with open(targetFile, 'w') as w:
    for s in targetScores:
      w.write(str(s) + '\n')
  
  fdp2, decoyFile = tempfile.mkstemp()
  with open(decoyFile, 'w') as w:
    for s in decoyScores:
      w.write(str(s) + '\n')
  
  fdq, qvalFile = tempfile.mkstemp()
  if tdcInput:
    tdcInputFlag = "-Y"
  else:
    tdcInputFlag = ""
  
  rc = subprocess.call("qvality %s %s %s -v %d > %s" % (tdcInputFlag, targetFile, decoyFile, VERB, qvalFile), shell=True)
  if includePEPs:
    qvals, peps = parseQvalues(qvalFile, includePEPs = includePEPs)
  else:
    qvals = parseQvalues(qvalFile)
  
  if includeDecoys:
    targetScores = np.array(targetScores[::-1])
    maxIdx = len(targetScores) - 1
    for s in decoyScores[::-1]:
      idx = np.searchsorted(targetScores, s, side='left')
      qvals.append(qvals[maxIdx-idx])
      if includePEPs:
        peps.append(peps[maxIdx-idx])
    qvals = sorted(qvals)
    if includePEPs:
      peps = sorted(peps)
      
  os.unlink(targetFile)
  os.unlink(decoyFile)
  os.unlink(qvalFile)
  if includePEPs:
    return qvals, peps
  else:
    return qvals
    
def parseQvalues(qvalFile, includePEPs = False):
  reader = csv.reader(open(qvalFile, 'r'), delimiter = '\t')
  next(reader)
  
  qvals, peps = list(), list()
  for row in reader:
    if includePEPs:
      peps.append(float(row[1]))
    qvals.append(float(row[2]))
  
  if includePEPs:
    return qvals, peps
  else:
    return qvals

def countBelowFDR(peps, qvalThreshold):
  return np.count_nonzero(np.cumsum(peps) / np.arange(1, len(peps) + 1) < qvalThreshold)

def getPEPFromScoreLambda(targetScores, decoyScores):
  if len(decoyScores) == 0:
    sys.exit("ERROR: No decoy hits found, check if the correct decoy prefix was specified with the --decoy_pattern flag")
  
  targetScores = np.array(targetScores)
  decoyScores = np.array(decoyScores)
  _, peps = getQvaluesFromScores(targetScores, decoyScores, includePEPs = True, includeDecoys = True, tdcInput = True)
  
  print("  Identified", countBelowFDR(peps, 0.01), "PSMs at 1% FDR")
  
  peps = peps[::-1] # PEPs in descending order, highest PEP first
  
  allScores = np.concatenate((targetScores, decoyScores))
  allScores.sort()  # scores in ascending order, lowest score first
  getPEPFromScore = lambda score : peps[min(np.searchsorted(allScores, score, side = 'left'), len(peps) - 1)] if not np.isnan(score) else 1.0
  
  return getPEPFromScore

def fdrsToQvals(fdrs):
  qvals = [0] * len(fdrs)
  if len(fdrs) > 0:
    qvals[len(fdrs)-1] = fdrs[-1]
    for i in range(len(fdrs)-2, -1, -1):
      qvals[i] = min(qvals[i+1], fdrs[i])
  return qvals

def getPEPAtFDRThreshold(peps, qvalThreshold):
  qvals = np.divide(np.cumsum(peps), np.arange(1,len(peps)+1))
  if qvals[-1] > qvalThreshold:
    idx = np.argmax(qvals > qvalThreshold)
  else:
    idx = -1
  pepThreshold = peps[idx]
  return pepThreshold
   
def unitTestScoreInput():
  import scipy.stats
  import matplotlib.pyplot as plt
  import time
  
  for s in range(3,4):
    t0 = time.time()
    np.random.seed(s)
    N = 500000
    #targetScores = np.round(np.random.normal(2.0,1,N) + np.random.normal(0.0,1,N), 1)
    #decoyScores = np.round(np.random.normal(0.0,1,N), 1)
    targetScores = np.random.normal(2.0,1,N) + np.random.normal(0.0,1,N)
    decoyScores = np.random.normal(0.0,1,N)
    print("Generated input", time.time() - t0, "sec")
    
    t0 = time.time()
    _, peps1 = getQvaluesFromScores(targetScores, decoyScores, includePEPs = True, includeDecoys = True, tdcInput = True, plotRegressionCurve = False)
    print("Python", time.time() - t0, "sec")
    
    t0 = time.time()
    _, peps2 = getQvaluesFromScoresQvality(targetScores, decoyScores, includePEPs = True, includeDecoys = True, tdcInput = True)
    print("C++", time.time() - t0, "sec")
    
    sortedScores = sorted(targetScores + decoyScores, reverse = True)
    plt.figure(s+1)
    plt.plot(scipy.stats.norm.pdf(sortedScores, 0.0,1) / (scipy.stats.norm.pdf(sortedScores, 2.0,1) + scipy.stats.norm.pdf(sortedScores, 0.0,1)), label = 'Ground truth')
    plt.plot(peps1, label = 'Python')
    plt.plot(peps2, label = 'C++')
    plt.legend()
  plt.show()

def unitTestPvalInput():
  import scipy.stats
  import matplotlib.pyplot as plt
  
  for s in range(3,4):
    np.random.seed(s)
    N = 5000
    targetScores = np.round(np.concatenate((np.random.uniform(0,1,N), scipy.stats.truncnorm.rvs(0.0,1,loc=0,scale=0.5,size=N))), 8)
    _, peps1 = getQvaluesFromPvalues(targetScores, includePEPs = True)
    _, peps2 = getQvaluesFromPvaluesQvality(targetScores, includePEPs = True)
    
    plt.figure(s+1)
    targetScores = sorted(targetScores)
    plt.plot(scipy.stats.uniform.pdf(targetScores, 0,1) / (scipy.stats.truncnorm.pdf(targetScores, 0.0,1,loc=0,scale=0.5) + scipy.stats.uniform.pdf(targetScores, 0,1)), label = 'Ground truth')
    plt.plot(peps1, label = 'Python')
    plt.plot(peps2, label = 'C++')
    plt.legend()
  plt.show()

if __name__ == "__main__":
  unitTestScoreInput()
  unitTestPvalInput()
